package dev;

import infomat.algorithms.grouping.*;
import infomat.bridges.tct.algorithms.grouping.*;
import infomat.evaluation.*;
import infomat.io.*;
import infomat.io.textresult.*;
import infomat.vectorspace.*;
import infomat.vectorspace.similarity.*;
import infomat.vectorspace.weighting.*;

import mro.util.*;
import mro.util.properties.*;
import mro.util.measures.*;

import java.io.*;
import java.util.Iterator;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;
import java.util.Collections;

import org.xml.sax.SAXException;


/**
 *
 * Class for running clustering experiments.
 * Only works on sparse data, with such similarity measures.
 * Reads Properties (one for this class, one for each class it uses)
 * from a path hierarchy and generates results 
 * in a corresponding hiererarchy.
 *
 *
 * <br>
 * <br>
 * <code>java -cp classes/ -Xms1024m -Xmx1024m dev.ExperimentatorNumberOfWords</code>
 *
 * @version          081001
 * @author           Magnus Rosell
 *
 */
public class ExperimentatorNumberOfWords implements HasProperties{

    private File propertiesDirectory=null;
    private Properties properties = null;

    public ExperimentatorNumberOfWords(File propertiesDirectory){
	this.propertiesDirectory = propertiesDirectory;
	this.properties = getDefaultProperties();
    }

    public Properties getDefaultProperties(){

	Properties properties = 
	    new Properties("ExperimentatorNumberOfWords Properties", CurrentClassGetter.getClassName());
		
	try{
			
	    /****************/
	    Property p = new Property("Result Path", null, "Result Path");
	    properties.addProperty(p);

	    /****************/
	    p = new Property("Number of Repetitions", 10, "Number of times the clustering is done for each set of properties.");
	    properties.addProperty(p);

	    /****************/
	    p = new Property("Stoplist", null, "File with stopobjects (words for instance). Opposite dimension to \"Rows as Matrix Dimension\". Combined with stoplist Properties from file. Won't be used without such a Properties!");
	    properties.addProperty(p);

	    /****************/
	    p = new Property("Rows as Matrix Dimension", true, "Whether rows or columns are regarded. Overrides the similarity measures dimension.");
	    properties.addProperty(p);

	    /****************/
	    PropertyGroup pg = new PropertyGroup("Matrix Source");
	    p = new Property("Matrix File", null, "Matrix File");
	    pg.addProperty(p);
	    p = new Property("Token File", null, "Token File");
	    pg.addProperty(p);
	    p = new Property("Inpath", null, "Path from where files are read recursively");
	    pg.addProperty(p);
	    properties.addProperty(pg);

	    /****************/
	    pg = new PropertyGroup("Comparison Grouping");
	    p = new Property("IObjectGrouping File", null, "IObjectGrouping File");
	    pg.addProperty(p);

	    p = new Property("LocationGrouper", false, "Determine by location.");
	    pg.addProperty(p);
	    properties.addProperty(pg);


	}catch(PropertyNameConflictException pnce){
	    // Won't happen here
	}
	return properties;
    }


    public Object getPropertyValue(String name){
    	return properties.getPropertyValue(name);
    }
    public void setProperty(String name, Object value) throws PropertyNotFoundException{
    	properties.setProperty(name, value);
    }

    public mro.util.properties.Properties getProperties(){
    	return properties;
    }

    public void setProperties(mro.util.properties.Properties properties){
    	this.properties=properties;
    }



    //    private static PropertiesPool pool = null;

    public static void main(String[] args) throws IOException, PropertyNotFoundException{
	String usageString = "java infomat.Experimentator <properties path>";
	File dir = Clusterer.checkInputReturnDirectory(args, usageString);

	PropertiesPool pool = new PropertiesPool();
	try{
	    pool.loadPropertiesInDirectory(dir);
	}catch(org.xml.sax.SAXException saxe){
	    badInput("Bad xml in one Properties, or other SAX-problem!");
	}

	System.out.println(dir);

	ExperimentatorNumberOfWords experimentator = new ExperimentatorNumberOfWords(dir);
	Properties p = pool.getPropertiesForObject(experimentator);
	if(p==null){
	    badInput("No Experimentator Properties!");
	}
	Clusterer.checkClustererProperties(p); // same properties to test!!!
	experimentator.setProperties(p);

	experimentator.doIt();

    }

    private static void badInput(String comment){
	System.out.println();
	System.out.println("Bad input!");
	System.out.println(comment);
	System.exit(0);
    }

    private void init(){

	// Load Original Matrix
	originalMatrix = Clusterer.loadMatrix(getProperties());

	// Start with the original matrix
	matrix = new IMatrix(originalMatrix);
	poolForMatrix = new PropertiesPool();

	// Make reference grouping
	referenceGrouping = 
	    Clusterer.constructReferenceGrouping(getProperties(), matrix);

	// What matrix dimension?
	rows = (Boolean) getPropertyValue("Rows as Matrix Dimension");
    }


    private boolean rows;
    private IMatrix originalMatrix;

    private void doIt() throws IOException, PropertyNotFoundException{

	init();

	PropertiesPool pool = new PropertiesPool();
	recurseExperimentDirectory(propertiesDirectory, pool);
    }


    private void recurseExperimentDirectory(File dir, PropertiesPool pool)
	throws IOException{ //, SAXException{
	
	System.out.println("recurseExperimentPath(): "+dir.getPath());

	
	pool = accumulateToPool(pool, dir);
	//	preprocess(pool);

	boolean directories=false;
	File[] list = dir.listFiles();
	for(int i=0; i<list.length; i++){
	    if(list[i].isDirectory()){
		directories=true;
	    }
	}
	if(directories){
	    // Continue recursing
	    for(int i=0; i<list.length; i++){
		if(list[i].isDirectory()){
		    recurseExperimentDirectory(list[i], pool);
		}
	    }
	}else{
	    // Do an experiment
	    
	    doOneExperiment(pool, dir);
	}

    }

    private PropertiesPool accumulateToPool(PropertiesPool pool, File dir)
	throws IOException{

	PropertiesPool retPool = new PropertiesPool(pool);

	try{
	    retPool.loadPropertiesInDirectory(dir);
	}catch(org.xml.sax.SAXException saxe){
	    badInput("Bad xml in one Properties, or other SAX-problem!");
	}

	return retPool;
    }


    private void doOneExperiment(PropertiesPool pool, File dir)
	throws IOException{//, PropertyNotFoundException{

	String topPropertiesPath = PathHelper.convertToCurrentFileSeparator(propertiesDirectory.getCanonicalPath());
	String propertiesPath = PathHelper.convertToCurrentFileSeparator(dir.getCanonicalPath());
	String propertiesPathDifference = 
	    PathHelper.overshoootForSecond(topPropertiesPath, propertiesPath);

	String resultPath = PathHelper.convertToCurrentFileSeparator((String) getPropertyValue("Result Path"));
	resultPath = resultPath + File.separator + propertiesPathDifference;

	LocationGrouper locationGrouper = new LocationGrouper();


	int numberOfTimes = 
	    (Integer) getPropertyValue("Number of Repetitions");

	constructLimits();


	for(int reductionType=0; reductionType<2; reductionType++){

	IMatrix reducingMatrix = new IMatrix(originalMatrix);
	minCounter = 0;	
	maxCounter = 0;	

	matrix = new IMatrix(reducingMatrix);
	preprocess(pool, matrix);


	boolean goOn=true;
	int experimentNumber = 10;
	while(goOn){
	    
	    SparseISimilarity sparseISimilarity = 
		Clusterer.constructSimilarityMeasure(pool, matrix, rows);
	    
	    GroupingAlgorithm ga = 
		Clusterer.getAlgorithm(pool, matrix, sparseISimilarity);
	    RandomClusterer rnd = new RandomClusterer();
	    try{
		int numClusters = (Integer) 
		    ga.getPropertyValue("Number of clusters");
		rnd.setProperty("Number of clusters", numClusters);
	    }catch(mro.util.properties.PropertyNotFoundException pnfe){
		// Ska inte hnda
		System.out.println("WTF!");
		System.exit(0);
	    }


	    IObjectSet set = matrix.getColumnIObjectSet();
	    if(rows)
		set = matrix.getRowIObjectSet();
	    

	    for(int i=0; i<numberOfTimes; i++){
		IObjectGrouping clustering = ga.cluster(set);
		long time = System.currentTimeMillis();
		
		String tmpResultPath = resultPath;
		if(reductionType==0)
		    tmpResultPath += "/minReduction/";
		if(reductionType==1)
		    tmpResultPath += "/maxReduction/";
		tmpResultPath += ""+experimentNumber;

		Clusterer.evaluate(pool, matrix, clustering, referenceGrouping, sparseISimilarity, tmpResultPath+"/clustering/", time);

		IObjectGrouping rndClustering = rnd.cluster(set);
		Clusterer.evaluate(pool, matrix, rndClustering, referenceGrouping, sparseISimilarity, tmpResultPath+"/randomClustering/", time);
	    }



	    matrix = new IMatrix(reducingMatrix);

	    if(reductionType==0)
		goOn = minReduction(matrix);
	    if(reductionType==1)
		goOn = maxReduction(matrix);

	    preprocess(pool, matrix);


	    experimentNumber++;
	}


	}

	File pPath = 
	    new File(resultPath+File.separator+"properties"+File.separator);
	if(!pPath.exists())
	    pPath.mkdirs();

	pool.savePropertiesToDirectory(pPath);

    }

    //    private HashMap<IObject, Integer> numOfTextsMap = 
    //	new HashMap<IObject, Integer>();

    //    private Vector<Integer> minLimits = new Vector<Integer>();
    //    private Vector<Integer> maxLimits = new Vector<Integer>();

    Vector<HashSet<IObject>> minRemoves = new Vector<HashSet<IObject>>();
    Vector<HashSet<IObject>> maxRemoves = new Vector<HashSet<IObject>>();


    private void constructLimits(){
	Vector<IObject> vec = 
	    originalMatrix.getColumnIObjectSet().getOrdered();

	HashMap<Integer, HashSet<IObject>> textOccurrenceMap = 
	    new HashMap<Integer, HashSet<IObject>>();

	int min=Integer.MAX_VALUE;
	int max=0;
	for(int i=0; i<vec.size(); i++){
	    IObject io = vec.get(i);
	    HashSet set = 
		originalMatrix.getRowIMatrixCellsForColumnObject(io);

	    int size = set.size();
	    if(size>max)
		max=size;
	    if(size<min)
		min=size;

	    HashSet<IObject> ioset = textOccurrenceMap.get(size);
	    if(ioset == null){
		ioset = new HashSet<IObject>();
		textOccurrenceMap.put(size, ioset);
	    }
	    ioset.add(io);

	    //	    numOfTextsMap.put(io, size);
	}

	System.out.print("maxDocFreq = ["+max+", ");
	int limit = max/2;
	while(limit>5){
	//	while(limit>2){
	    System.out.print(limit+", ");

	    HashSet<IObject> ioset = new HashSet<IObject>();
	    for(int i=limit; i<=max; i++){
		HashSet<IObject> ioset2 = textOccurrenceMap.get(i);
		if(ioset2!=null)
		    ioset.addAll(ioset2);
	    }
	    maxRemoves.add(ioset);
	    limit /= 2;
	}
	System.out.println("]");


	System.out.print("minDocFreq = ["+min+", ");
	limit = min*2;
	while(limit<max/5){
	//	while(limit<max/2){
	    System.out.print(limit+", ");

	    HashSet<IObject> ioset = new HashSet<IObject>();
	    for(int i=limit; i>=min; i--){
		HashSet<IObject> ioset2 = textOccurrenceMap.get(i);
		if(ioset2!=null)
		    ioset.addAll(ioset2);
	    }
	    minRemoves.add(ioset);
	    limit *= 2;
	}
	System.out.println("]");

	System.out.println("maxRemoves: "+maxRemoves.size());
	System.out.println("minRemoves: "+minRemoves.size());

    }




    private int minCounter = 0;
    private boolean minReduction(IMatrix currentMatrix){
	if(minCounter>minRemoves.size()-1)
	    return false;

	IObjectSet rowSet = currentMatrix.getRowIObjectSet();
	IObjectSet colSet = currentMatrix.getColumnIObjectSet();
	HashSet<IObject> rmSet = minRemoves.get(minCounter);
	Iterator<IObject> it=rmSet.iterator();
	while(it.hasNext()){
	    IObject io = it.next();
	    colSet.remove(io);
	}
	currentMatrix.purgeMatrixToGroups(rowSet, colSet);

	minCounter++;
	return true;
    }

    private int maxCounter = 0;
    private boolean maxReduction(IMatrix currentMatrix){
	if(maxCounter>maxRemoves.size()-1)
	    return false;

	IObjectSet rowSet = currentMatrix.getRowIObjectSet();
	IObjectSet colSet = currentMatrix.getColumnIObjectSet();
	HashSet<IObject> rmSet = maxRemoves.get(maxCounter);
	Iterator<IObject> it=rmSet.iterator();
	while(it.hasNext()){
	    IObject io = it.next();
	    colSet.remove(io);
	}
	currentMatrix.purgeMatrixToGroups(rowSet, colSet);

	maxCounter++;
	return true;
    }



    private IMatrix matrix;
    private PropertiesPool poolForMatrix;
    private IObjectGrouping referenceGrouping;
    
    private void preprocess(PropertiesPool pool, IMatrix originalMatrix){

	System.out.println("preprocess");

	//	if(poolsDifferentInPreprocessing(pool, poolForMatrix)){
	    matrix = new IMatrix(originalMatrix);
	    poolForMatrix = pool;

	    System.out.println(" - matrix: "+matrix.getSize());

	    Clusterer.stoplist(getProperties(), pool, matrix);
	    Clusterer.filter(pool, matrix);
	    weight(pool, matrix);

	    System.out.println(" - matrix: "+matrix.getSize());

	    // Make reference grouping
	    referenceGrouping = 
		Clusterer.constructReferenceGrouping(getProperties(), matrix);
	    //	}

	    

    }

    private boolean 
	poolsDifferentInPreprocessing(PropertiesPool pool, 
				      PropertiesPool poolForMatrix){

	Stoplist sl = new Stoplist();
	Properties ps1 = pool.getPropertiesForObject(sl);
	Properties ps2 = poolForMatrix.getPropertiesForObject(sl);
	if(ps1!=null && ps2!=null && !ps1.equals(ps2))
	    return true;
	if((ps1==null && ps2!=null) || (ps1!=null && ps2==null))
	    return true;

	IMatrixCellFilter matrixCellFilter = new IMatrixCellFilter();
	ps1 = pool.getPropertiesForObject(matrixCellFilter);
	ps2 = poolForMatrix.getPropertiesForObject(matrixCellFilter);
	if(ps1!=null && ps2!=null && !ps1.equals(ps2))
	    return true;
	if((ps1==null && ps2!=null) || (ps1!=null && ps2==null))
	    return true;

	ps1=null;
	ps2=null;
	Vector<Weighter> weighters1 = Clusterer.getWeighters(pool);
	if(weighters1!=null){
	    if(weighters1.size()>1)
		badInput("Many Weighters!");
	    ps1 = weighters1.get(0).getProperties();
	}
	Vector<Weighter> weighters2 = Clusterer.getWeighters(poolForMatrix);
	if(weighters2!=null){
	    if(weighters2.size()>1)
		badInput("Many Weighters!");
	    ps2 = weighters2.get(0).getProperties();
	}
	if(ps1!=null && ps2!=null && !ps1.equals(ps2))
	    return true;
	if((ps1==null && ps2!=null) || (ps1!=null && ps2==null))
	    return true;

	return false;
    }

    private static String weight(PropertiesPool pool, IMatrix matrix){
	Vector<Weighter> weighters = Clusterer.getWeighters(pool);
	if(weighters==null){
	    // Do nothing
	    return null;
	}
	if(weighters.size()>1){
	    //	    return "Many Weighters!";
	    badInput("Many Weighters!");
	}

	Weighter weighter = weighters.elementAt(0);
	weighter.weight(matrix);

	return null; // Everything is alright!
    }



}
