package infomat;

import infomat.algorithms.grouping.*;
import infomat.bridges.tct.algorithms.grouping.*;
import infomat.evaluation.*;
import infomat.io.*;
import infomat.io.textresult.*;
import infomat.io.textword.*;
import infomat.vectorspace.*;
import infomat.vectorspace.similarity.*;
import infomat.vectorspace.weighting.*;

import mro.util.*;
import mro.util.properties.*;
import mro.util.measures.*;

import java.io.*;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;

import org.xml.sax.SAXException;


/**
 *
 * Class for running clustering tasks without the GUI.
 * Only works on sparse data, with such similarity measures.
 * Reads Properties (one for this class, one for each class it uses)
 * from a directory. Only one of each Properties is
 * allowed. 
 *
 *
 * <br>
 * <br>
 * <code>java -cp classes/ -Xms1024m -Xmx1024m infomat.JohanClusterer</code>
 *
 * @version          081001
 * @author           Magnus Rosell
 *
 */
public class JohanClusterer implements HasProperties{

    private Properties properties = null;
    private PropertiesPool pool;

    public JohanClusterer(PropertiesPool pool){
	this.pool = pool;
	this.properties = getDefaultProperties();
    }

    public Properties getDefaultProperties(){

	Properties properties = 
	    new Properties("Johan Clusterer Properties", CurrentClassGetter.getClassName());
		
	try{
			
	    /****************/
	    Property p = new Property("Result Path", null, "Result Path");
	    properties.addProperty(p);

	    /****************/
	    p = new Property("Stoplist", null, "File with stopobjects (words for instance). Opposite dimension to \"Rows as Matrix Dimension\". Combined with stoplist Properties from file.");
	    properties.addProperty(p);

	    /****************/
	    p = new Property("Rows as Matrix Dimension", true, "Whether rows or columns are regarded. Overrides the similarity measures dimension.");
	    properties.addProperty(p);

	    /****************/
	    PropertyGroup pg = new PropertyGroup("Matrix Source");
	    p = new Property("Matrix File", null, "Matrix File");
	    pg.addProperty(p);
	    p = new Property("Token File", null, "Token File");
	    pg.addProperty(p);
	    p = new Property("Inpath", null, "Path from where files are read recursively");
	    pg.addProperty(p);
	    properties.addProperty(pg);

	    /****************/
	    pg = new PropertyGroup("Comparison Grouping");
	    p = new Property("IObjectGrouping File", null, "IObjectGrouping File");
	    pg.addProperty(p);

	    p = new Property("LocationGrouper", false, "Determine by location.");
	    pg.addProperty(p);
	    properties.addProperty(pg);


	}catch(PropertyNameConflictException pnce){
	    // Won't happen here
	}
	return properties;
    }


    public Object getPropertyValue(String name){
    	return properties.getPropertyValue(name);
    }
    public void setProperty(String name, Object value) throws PropertyNotFoundException{
    	properties.setProperty(name, value);
    }

    public mro.util.properties.Properties getProperties(){
    	return properties;
    }

    public void setProperties(mro.util.properties.Properties properties){
    	this.properties=properties;
    }



    public static void main(String[] args) throws IOException, PropertyNotFoundException{
	String usageString = "java infomat.JohanClusterer <properties path>";
	File dir = checkInputReturnDirectory(args, usageString);

	PropertiesPool pool = new PropertiesPool();
	try{
	    pool.loadPropertiesInDirectory(dir);
	}catch(org.xml.sax.SAXException saxe){
	    badInput("Bad xml in one Properties, or other SAX-problem!");
	}

	JohanClusterer clusterer = new JohanClusterer(pool);
	Properties p = pool.getPropertiesForObject(clusterer);
	if(p==null){
	    badInput("No Johan Clusterer Properties!");
	}
	checkClustererProperties(p);
	clusterer.setProperties(p);

	clusterer.doIt();

	/*
	Clusterer clusterer = new Clusterer();
	clusterer.doIt(args);
	*/
    }

    private static void badInput(String comment){
	System.out.println();
	System.out.println("Bad input!");
	System.out.println(comment);
	System.exit(0);
    }


    protected static void checkClustererProperties(Properties properties){
	
	// Result path
	String resultFileName = (String) properties.getPropertyValue("Result Path");
	if(resultFileName==null)
	    badInput("No result path!");
	File dir = new File(resultFileName);
	if(!dir.exists())
	    dir.mkdirs();

	// Indata
	String matrixFileName = (String) properties.getPropertyValue("Matrix File");
	String tokenFileName = (String) properties.getPropertyValue("Token File");
	String inpathName = (String) properties.getPropertyValue("Inpath");
	int number = 0;
	if(matrixFileName!=null)
	    number++;
	if(tokenFileName!=null)
	    number++;
	if(inpathName!=null)
	    number++;

	if(number==0)
	    badInput("No data!");
	if(number!=1)
	    badInput("To many data files (matrix, token, inpath)!");

	// Reference
	String grouping = (String) properties.getPropertyValue("IObjectGrouping File");
	boolean locationGrouper = (Boolean) properties.getPropertyValue("LocationGrouper");
	if(grouping!=null && locationGrouper)
	    badInput("To many reference groupings!");
    }
    
    protected static File checkInputReturnDirectory(String[] args, String usageString) throws IOException, PropertyNotFoundException{

	if(args.length!=1){
	    badInput(usageString);
	}

	File dir = new File(args[0]);
	if(!dir.isDirectory())
	    badInput("Properties: not a directory: "+args[0]);

	return dir;
    }


    public void doIt() throws IOException, PropertyNotFoundException{

	// Matrix
	IMatrix matrix = loadMatrix(getProperties());
	boolean rows = (Boolean) getPropertyValue("Rows as Matrix Dimension");

	// Preprocess	
	stoplist(getProperties(), pool, matrix);
	filter(pool, matrix);
	weight(pool, matrix);
	
	// Reference
	IObjectGrouping referenceGrouping = 
	    constructReferenceGrouping(getProperties(), matrix);

	// Similarity
	SparseISimilarity sparseISimilarity = 
	    constructSimilarityMeasure(pool, matrix, rows);

	// Some info
	String resultPath = 
	    (String) properties.getPropertyValue("Result Path");
	long time = System.currentTimeMillis();


	// Cluster
	GroupingAlgorithm ga = getAlgorithm(pool, matrix, sparseISimilarity);
	IObjectSet set = matrix.getColumnIObjectSet();
	if(rows)
	    set = matrix.getRowIObjectSet();
	//	IObjectGrouping clustering = ga.cluster(set);
	IObjectGrouping clustering = cluster(ga, set, pool, matrix, referenceGrouping, sparseISimilarity, resultPath, time);


	// Write Result and Evaluation
	writeTextClusteringResult(pool, matrix, clustering, sparseISimilarity, resultPath, time);

	writeFullClusteringResult(pool, matrix, clustering, sparseISimilarity, resultPath, time);

	Measures measures = evaluate(pool, matrix, clustering, referenceGrouping, sparseISimilarity, resultPath, time);

	String evaluationFileName = resultPath;
	evaluationFileName = evaluationFileName + File.separator + 
	    "clusteringEvaluation."+time+".xml";

	measures.save(evaluationFileName);

    }

    
    private IObjectGrouping cluster(GroupingAlgorithm ga, IObjectSet set, PropertiesPool pool, IMatrix matrix, IObjectGrouping referenceGrouping, SparseISimilarity sparseISimilarity, String resultPath, long time){


	IObjectGrouping bestIOGING = ga.cluster(set);
	Measures measures = evaluate(pool, matrix, bestIOGING, referenceGrouping, sparseISimilarity, resultPath, time);
	MeasureGroup mg = measures.getMeasureGroup("Intrinsic Measures");
	mg = mg.getMeasureGroup("Evaluated Grouping");
	Measure m = mg.getMeasure("Weighted Avereage Self Similarity");
	double bestAvgSim = m.getValue();

	for(int i=0; i<9; i++){
	    IObjectGrouping ioging = ga.cluster(set);
	    measures = evaluate(pool, matrix, ioging, referenceGrouping, sparseISimilarity, resultPath, time);
	    mg = measures.getMeasureGroup("Intrinsic Measures");
	    mg = mg.getMeasureGroup("Evaluated Grouping");
	    m = mg.getMeasure("Weighted Avereage Self Similarity");
	    double iogingAvgSim = m.getValue();
	    if(iogingAvgSim>bestAvgSim){
		bestAvgSim = iogingAvgSim;
		bestIOGING = ioging;
	    }
	}

	return bestIOGING;
    }




    protected static IMatrix loadMatrix(Properties properties){
	
	boolean rows = (Boolean) 
	    properties.getPropertyValue("Rows as Matrix Dimension");

	String matrixFileName = (String) 
	    properties.getPropertyValue("Matrix File");
	String tokenFileName = (String) 
	    properties.getPropertyValue("Token File");
	String inpathName = (String) 
	    properties.getPropertyValue("Inpath");

	IMatrix matrix = null;

	if(matrixFileName!=null){
	    File matrixFile = new File(matrixFileName);
	    IMatrixIO iomatrix = new IMatrixIO();
	    matrix = iomatrix.load(matrixFile);
	}else if(tokenFileName!=null){
	    File tokenFile = new File(tokenFileName);
	    TokenFileTextWordIO tfio = new TokenFileTextWordIO();
	    matrix = tfio.load(tokenFile);
	}else if(inpathName!=null){
	    File inpathFile = new File(inpathName);
	    TextWordDirectoryReader tdr = new TextWordDirectoryReader();
	    matrix = tdr.load(inpathFile);
	}
	return matrix;
    }

    protected static IObjectGrouping 
	constructReferenceGrouping(Properties props, IMatrix matrix){

	IObjectGrouping referenceGrouping = null;

	boolean rows = (Boolean) 
	    props.getPropertyValue("Rows as Matrix Dimension");
	String groupingFileName = (String) 
	    props.getPropertyValue("IObjectGrouping File");
	boolean useLocationGrouper = (Boolean) 
	    props.getPropertyValue("LocationGrouper");
	
	if(groupingFileName!=null && useLocationGrouper)
	    badInput("Two reference groupings!");

	if(groupingFileName!=null){
	    File file = new File(groupingFileName);
	    IObjectGroupingIO iobio = new IObjectGroupingIO();
	    referenceGrouping = iobio.load(rows, matrix, file);
	}else if(useLocationGrouper){
	    LocationGrouper locationGrouper = new LocationGrouper();
	    IObjectSet set = matrix.getRowIObjectSet();
	    if(!rows)
		set = matrix.getColumnIObjectSet();
	    referenceGrouping = 
		locationGrouper.cluster(set);
	}

	return referenceGrouping;
    }


    protected static void stoplist(Properties properties, PropertiesPool pool, IMatrix matrix){
	boolean stop = false;

	Stoplist sl = new Stoplist();
	Properties p = pool.getPropertiesForObject(sl);
	if(p!=null){
	    sl.setProperties(p);
	    stop=true;
	}
	
	String stoplistFileName = 
	    (String) properties.getPropertyValue("Stoplist");
	if(stoplistFileName!=null){
	    sl.readFile(stoplistFileName);
	}

	if(stop){
	    boolean rows = (Boolean) properties.getPropertyValue("Rows as Matrix Dimension");
	    sl.removeIObjectsInStoplist(matrix, !rows);
	}
    }

    protected static void filter(PropertiesPool pool, IMatrix matrix){
	boolean doFilter = false;
	IMatrixCellFilter matrixCellFilter = new IMatrixCellFilter();
	Properties p = pool.getPropertiesForObject(matrixCellFilter);
	if(p!=null){
	    matrixCellFilter.setProperties(p);
	    doFilter = true;
	}
	if(doFilter)
	    matrixCellFilter.filter(matrix);
    }


    protected static Vector<Weighter> getWeighters(PropertiesPool pool){
	Vector<Properties> props = 
	    pool.getPropertiesWithSuperClassName("infomat.vectorspace.weighting.Weighter");
	if(props==null || props.size()==0)
	    return null;
	
	Vector<Weighter> weighters = new Vector<Weighter>();

	try{
	    for(int i=0; i<props.size(); i++){
		Properties p = props.elementAt(i);
		String className = p.getOwnerClassName();
		
		Class c = Class.forName(className);
		Weighter weighter = (Weighter) c.newInstance();
		weighter.setProperties(p);

		weighters.add(weighter);
	    }

	}catch(ClassNotFoundException cnfe){
	    System.out.println("Problems with weighting!");
	    cnfe.printStackTrace();
	    System.exit(0);
	}catch(InstantiationException ie){
	    System.out.println("Problems with weighting!");
	    ie.printStackTrace();
	    System.exit(0);
	}catch(IllegalAccessException iae){
	    System.out.println("Problems with weighting!");
	    iae.printStackTrace();
	    System.exit(0);
	}

	return weighters;
    }


    protected static String weight(PropertiesPool pool, IMatrix matrix){
	Vector<Weighter> weighters = getWeighters(pool);
	if(weighters==null){
	    //	    return "No Weighter!";
	    badInput("No Weighter!");
	}
	if(weighters.size()>1){
	    //	    return "Many Weighters!";
	    badInput("Many Weighters!");
	}

	Weighter weighter = weighters.elementAt(0);
	weighter.weight(matrix);

	return null; // Everything is alright!
    }


    private static Vector<SparseISimilarity> getSimilarityMeasures(PropertiesPool pool, IMatrix matrix, boolean rows){
	Vector<Properties> props = 
	    pool.getPropertiesWithSuperClassName("infomat.vectorspace.similarity.SparseISimilarity");

	if(props==null || props.size()==0)
	    return null;

	Vector<SparseISimilarity> similarities = 
	    new Vector<SparseISimilarity>();

	try{
	    for(int i=0; i<props.size(); i++){
	    
		Properties p = props.elementAt(i);
		String className = p.getOwnerClassName();
		
		Class c = Class.forName(className);
		SparseISimilarity sparseISimilarity = 
		    (SparseISimilarity) c.newInstance();
		sparseISimilarity.setProperties(p);
		
		if(sparseISimilarity instanceof NeedsIMatrix)
		    ((NeedsIMatrix) sparseISimilarity).setIMatrix(matrix);
		if(sparseISimilarity instanceof NeedsIMatrixDimension)
		    ((NeedsIMatrixDimension) sparseISimilarity).setIMatrixDimension(rows);
		similarities.add(sparseISimilarity);

	    }

	}catch(ClassNotFoundException cnfe){
	    System.out.println("Problems with similarity measure!");
	    cnfe.printStackTrace();
	    System.exit(0);
	}catch(InstantiationException ie){
	    System.out.println("Problems with similarity measure!");
	    ie.printStackTrace();
	    System.exit(0);
	}catch(IllegalAccessException iae){
	    System.out.println("Problems with similarity measure!");
	    iae.printStackTrace();
	    System.exit(0);
	}

	return similarities;

    }

    protected static SparseISimilarity 
	constructSimilarityMeasure(PropertiesPool pool, 
				   IMatrix matrix, boolean rows){

	Vector<SparseISimilarity> similarities = 
	    getSimilarityMeasures(pool, matrix, rows);

	if(similarities==null)
	    badInput("No Similarity Measure!");
	if(similarities.size()>1)
	    badInput("Many Similarity Measures!");
	    
	return similarities.elementAt(0);
    }



    protected static GroupingAlgorithm getAlgorithm(PropertiesPool pool, IMatrix matrix, SparseISimilarity sparseISimilarity){
	
	Vector<GroupingAlgorithm> gas = 
	    getAlgorithms(pool, matrix, sparseISimilarity);
	if(gas==null || gas.size()==0)
	    badInput("No Clustering Algorithm!");
	if(gas.size()>1)
	    badInput("Many Clustering Algorithms!");

	return gas.elementAt(0);
    }


    private static Vector<GroupingAlgorithm> getAlgorithms(PropertiesPool pool, IMatrix matrix, SparseISimilarity sparseISimilarity){
	Vector<Properties> props = 
	    pool.getPropertiesWithSuperClassName("infomat.algorithms.grouping.GroupingAlgorithm");
	if(props==null || props.size()==0)
	    return null;
	
	Vector<GroupingAlgorithm> gas = 
	    new Vector<GroupingAlgorithm>();


	try{
	    for(int i=0; i<props.size(); i++){

		Properties p = props.elementAt(i);
		String className = p.getOwnerClassName();
		Class c = Class.forName(className);
		GroupingAlgorithm ga = (GroupingAlgorithm) c.newInstance();
		ga.setProperties(p);
		
		if(ga instanceof NeedsIMatrix)
		    ((NeedsIMatrix) ga).setIMatrix(matrix);
		if(ga instanceof NeedsISimilarity){
		    ((NeedsISimilarity) ga).setSimilarity(sparseISimilarity);
		}
		if(ga instanceof NeedsSparseISimilarity){
		    ((NeedsSparseISimilarity) ga).setSimilarity(sparseISimilarity);
		}
		
		gas.add(ga);
	    }

	}catch(ClassNotFoundException cnfe){
	    System.out.println("Problems with clustering algorithm!");
	    cnfe.printStackTrace();
	    System.exit(0);
	}catch(InstantiationException ie){
	    System.out.println("Problems with clustering algorithm!");
	    ie.printStackTrace();
	    System.exit(0);
	}catch(IllegalAccessException iae){
	    System.out.println("Problems with clustering algorithm!");
	    iae.printStackTrace();
	    System.exit(0);
	}
	
	return gas;
    }



    protected static void writeFullClusteringResult(PropertiesPool pool, IMatrix matrix, IObjectGrouping clustering, SparseISimilarity sparseISimilarity, String resultPath, long time){

	IObjectGroupingIO iogingio = new IObjectGroupingIO();
	Properties p = pool.getPropertiesForObject(iogingio);
	if(p!=null){
	    iogingio.setProperties(p);
	}else{
	    return;
	}

	String resultFileName = resultPath;

	resultFileName = resultFileName + File.separator + 
	    "clustering."+time+".xml";

	File resFile = new File(resultFileName);
	iogingio.save(clustering, matrix, sparseISimilarity, resFile);
    }


    protected static void writeTextClusteringResult(PropertiesPool pool, IMatrix matrix, IObjectGrouping clustering, SparseISimilarity sparseISimilarity, String resultPath, long time){

	IObjectGroupingTextResult gtio = new IObjectGroupingTextResult();
	Properties p = pool.getPropertiesForObject(gtio);
	if(p!=null){
	    gtio.setProperties(p);
	}else{
	    return;
	}

	String resultFileName = resultPath;

	resultFileName = resultFileName + File.separator + 
	    "clusteringResult."+time+".xml";

	File resFile = new File(resultFileName);
	gtio.save(matrix, clustering, sparseISimilarity, resFile);
    }



    protected static Measures evaluate(PropertiesPool pool, IMatrix matrix, IObjectGrouping clustering, IObjectGrouping referenceGrouping, SparseISimilarity sparseISimilarity, String resultPath, long time){

	String evaluationFileName = resultPath;
	evaluationFileName = evaluationFileName + File.separator + 
	    "clusteringEvaluation."+time+".xml";

	Evaluation eval = 
	    new Evaluation(matrix, clustering, referenceGrouping, sparseISimilarity, ""+time);
	Properties p = pool.getPropertiesForObject(eval);
	if(p!=null){
	    eval.setProperties(p);
	}else{
	    return null;
	}

	Measures measures = new Measures("Measures");
	try{
	    measures = eval.calculateAllMeasures();
	}
	catch(MeasureNameConflictException mnce){
	    System.out.println("Measure Name Conflict!");
	}

	return measures;
    }


}
