package infomat.evaluation;

import java.util.HashMap;
import java.util.Vector;
import java.io.*;

import infomat.vectorspace.*;

import mro.util.*;
import mro.util.measures.*;
import mro.util.properties.*;
import mro.util.doublehashmap.DoubleHashMap;


/**
 * For calculating extrinsic quality measures.
 * 
 * @version          080326
 * @author           Magnus Rosell
 */
public class ExtrinsicEvaluation {

    private Properties properties;

    private IObjectGrouping evalGrouping;
    private IObjectGrouping refGrouping;
    private String stamp;

    private DoubleHashMap confusionMatrix;

    public ExtrinsicEvaluation(IObjectGrouping evalGrouping, 
			       IObjectGrouping refGrouping,
			       String stamp){
	this.evalGrouping = evalGrouping;
	this.refGrouping = refGrouping;
	this.stamp=stamp;
	this.properties = getDefaultProperties();
	constructConfusionMatrix();
    }

    public Properties getDefaultProperties(){

	Properties properties = 
	    new Properties("Extrinsic Evaluation Properties", CurrentClassGetter.getClassName());
		
	try{
	    Property p = new Property("Group Evaluation", true, "Evaluate groups individually");
	    PropertyValueDescription pvd = new PropertyValueDescription(true, "Do it.");
	    p.addValueDescription(pvd);
	    pvd = new PropertyValueDescription(false, "Do not do it.");
	    p.addValueDescription(pvd);
	    properties.addProperty(p);

	}catch(PropertyNameConflictException pnce){
	    // Won't happen here
	}
	return properties;
    }

    private void constructConfusionMatrix(){
	confusionMatrix = new DoubleHashMap();
	Vector<IObjectGroup> g1groups = evalGrouping.getGroups();
	Vector<IObjectGroup> g2groups = refGrouping.getGroups();

	for(int i = 0; i < g1groups.size(); i++){
	    IObjectGroup iog1 = g1groups.get(i);
	    for(int j = 0; j < g2groups.size(); j++){
		IObjectGroup iog2 = g2groups.get(j);
		confusionMatrix.put(iog1, iog2, iog1.numberOfSharedIObjects(iog2));
	    }
	}		
    }

    public Measures calculateAllMeasures() 
	throws MeasureNameConflictException{
	Measures measures = new Measures("Extrinsic Measures");
	measures.addMeasure(calculateGlobalMeasures());

	measures.addMeasure(calculateGlobalPairMeasures());

	measures.addMeasure(calculateReferenceGlobalMeasures());

	boolean groupEvaluation = (Boolean) 
	    getPropertyValue("Group Evaluation");
	if(groupEvaluation)
	    measures.addMeasure(calculateGroupMeasures());

	measures.setAllMeasuresStampsRecursively(stamp);

	return measures;
    }

    public MeasureGroup allMeasuresInAGroup() 
	throws MeasureNameConflictException{
	MeasureGroup mg = new MeasureGroup("Extrinsic Measures");
	mg.addMeasure(calculateGlobalMeasures());

	mg.addMeasure(calculateGlobalPairMeasures());

	mg.addMeasure(calculateReferenceGlobalMeasures());

	boolean groupEvaluation = (Boolean) 
	    getPropertyValue("Group Evaluation");
	if(groupEvaluation)
	    mg.addMeasure(calculateGroupMeasures());

	mg.setAllMeasuresStampsRecursively(stamp);

	return mg;
    }


    public MeasureGroup calculateBasicMeasures() throws MeasureNameConflictException{

	MeasureGroup mg = new MeasureGroup("Basic Measures");

	MeasureGroup mg2 = new MeasureGroup("Evaluated Grouping");
	Measure m = new Measure("Number Of IObjects", evalGrouping.numberOfIObjects(), "Number Of IObjects in IObjectGroup");
	mg2.addMeasure(m);
	m = new Measure("Number Of IObjectGroups", evalGrouping.numberOfIObjectGroups(), "Number Of IObjects in IObjectGroup");
	mg2.addMeasure(m);
	mg.addMeasure(mg2);


	mg2 = new MeasureGroup("Reference Grouping");
	m = new Measure("Number Of IObjects", refGrouping.numberOfIObjects(), "Number Of IObjects in IObjectGroup");
	mg2.addMeasure(m);
	m = new Measure("Number Of IObjectGroups", refGrouping.numberOfIObjectGroups(), "Number Of IObjects in IObjectGroup");
	mg2.addMeasure(m);

	mg.addMeasure(mg2);

	return mg;
    }


    public MeasureGroup calculateGlobalMeasures() throws MeasureNameConflictException{
		
	MeasureGroup mg = new MeasureGroup("Global Extrinsic Measures");
		
	Measure m = new Measure("Purity", (double) purity(), "Purity");

	//System.out.println(m.getValue());
	/*
	try{
	    m.printXML(new PrintWriter(System.out));
	}
	catch(IOException ioe){}
	*/
	mg.addMeasure(m);

	double entropy = entropy();
	m = new Measure("Entropy", (double) entropy, "Entropy");
	mg.addMeasure(m);

	/*
	double entropy2 = entropy2();
	m = new Measure("Entropy2", (double) entropy, "Entropy2");
	mg.addMeasure(m);
	*/

	double normEntropy = entropy/entropyForIObjectGroups(refGrouping);
	m = new Measure("Norm Entropy", normEntropy, "Normalized Entropy ( E/E(K) )");
	mg.addMeasure(m);

	normEntropy = entropy/MoreMath.log2(refGrouping.numberOfIObjectGroups());
	m = new Measure("Norm Entropy by Number of Groups", normEntropy, "Normalized Entropy by Group Sizes ( E/log(|K|) )");
	mg.addMeasure(m);



	m = new Measure("Mutual Information", (double) mutualInformation(), "Mutual Information");
	mg.addMeasure(m);

	m = new Measure("NMI", (double) normMutualInformation(), "Normalized Mutual Information ( MI/sqrt(E(C)*E(K)) )");
	mg.addMeasure(m);

	m = new Measure("NMI by Number of Groups", (double) groupsNormMutualInformation(), "Normalized Mutual Information by Number of Groups ( 2*MI / log(|C|*|K|) )");
	mg.addMeasure(m);


	/******************/
	
	return mg;
    }


    public MeasureGroup calculateReferenceGlobalMeasures() throws MeasureNameConflictException{
		
	MeasureGroup mg = new MeasureGroup("Global Extrinsic Reference Measures");
	double evalEntropy = entropyForIObjectGroups(evalGrouping);
	double refEntropy = entropyForIObjectGroups(refGrouping);
	int evalGroups = evalGrouping.numberOfIObjectGroups();
	int refGroups = refGrouping.numberOfIObjectGroups();

	Measure m = new Measure("MI Max", Math.sqrt(refEntropy*evalEntropy), "NMI Max");
	mg.addMeasure(m);

	m = new Measure("MI by Number of Groups Max", MoreMath.log2(refGroups*evalGroups)/2, "MI by Number of Groups Max");
	mg.addMeasure(m);

	m = new Measure("Evaluated Grouping Entropy", (double) evalEntropy, "Entropy for distribution IObjects over IObjectGroups in Evaluated Grouping");
	mg.addMeasure(m);

	m = new Measure("Reference Grouping Entropy", (double) refEntropy, "Entropy for distribution IObjects over IObjectGroups in Reference Grouping");
	mg.addMeasure(m);

	m = new Measure("Evaluated / Reference Entropy", (evalEntropy/refEntropy), "\"Evaluated Grouping Entropy\" divided by \"Reference Grouping Entropy\"");
	mg.addMeasure(m);

	m = new Measure("Reference / Evaluated Entropy", (refEntropy/evalEntropy), "\"Reference Grouping Entropy\" divided by \"Evaluated Grouping Entropy\"");
	mg.addMeasure(m);	

	/*
	mg2 = new MeasureGroup("Evaluated Grouping");
	m = new Measure("Grouping Entropy", (double) evalEntropy, "Entropy for distribution IObjects over IObjectGroups in Grouping");
	mg2.addMeasure(m);

	double normEvalEntropy = evalEntropy/MoreMath.log2(refGroups);
	m = new Measure("Norm Entropy to other Grouping", normEvalEntropy, "\"Grouping Entropy\" normalized to other Grouping");
	mg2.addMeasure(m);

	normEvalEntropy = evalEntropy/MoreMath.log2(evalGroups);
	m = new Measure("Norm Entropy to Max", normEvalEntropy, "\"Grouping Entropy\" normalized to Maximal Entropy for grouping of same number of groups.");
	mg2.addMeasure(m);

	mg.addMeasure(mg2);


	mg2 = new MeasureGroup("Reference Grouping");
	m = new Measure("Grouping Entropy", (double) refEntropy, "Entropy for distribution IObjects over IObjectGroups in Grouping");
	mg2.addMeasure(m);

	double normRefEntropy = refEntropy/MoreMath.log2(evalGroups);
	m = new Measure("Norm Entropy to other Grouping", normRefEntropy, "\"Grouping Entropy\" normalized to other Grouping");
	mg2.addMeasure(m);

	normRefEntropy = refEntropy/MoreMath.log2(refGroups);
	m = new Measure("Norm Entropy to Max", normRefEntropy, "\"Grouping Entropy\" normalized to Maximal Entropy for grouping of same number of groups.");
	mg2.addMeasure(m);


	mg.addMeasure(mg2);
	*/


	/******************/
	
	return mg;
    }

    //Entropy for distribution IObjects over IObjectGroups in IObjectGrouping
    public static double entropyForIObjectGroups(IObjectGrouping grouping){
	double entropy = 0.0;
	int size = grouping.numberOfIObjects();
	Vector<IObjectGroup> groups = grouping.getGroups();
	for(int i=0; i<groups.size(); i++){
	    double count = groups.get(i).numberOfIObjects();
	    if(count>0){
		double p = count/size;
		entropy -= p*MoreMath.log2(p);
	    }
	}
	return entropy;
    }


    public MeasureGroup calculateGlobalPairMeasures() throws MeasureNameConflictException{

	double numOfIOs = evalGrouping.numberOfIObjects();
	double totNumOfPairs = numOfIOs * (numOfIOs-1) / 2;
	
	/*
	double pairsSS=0;
	double pairsSD=0;
	double pairsDS=0;
	double pairsDD=0;

	Vector<IObjectGroup> g1groups = evalGrouping.getGroups();
	for(int i = 0; i < g1groups.size(); i++){
	    IObjectGroup iog1 = g1groups.get(i);
	    Vector<IObjectGroup> g2groups = refGrouping.getGroups();
	    for(int j = 0; j < g2groups.size(); j++){
		IObjectGroup iog2 = g2groups.get(j);
		double count = 
		    (Integer) confusionMatrix.get(iog1, iog2);

		pairsSS += count * (count-1);
		pairsSD += count * 
		    (iog1.numberOfIObjects() - count);
		pairsDS += count * 
		    (iog2.numberOfIObjects() - count);
		pairsDD += count * 
		    (numOfIOs 
		     - iog1.numberOfIObjects()
		     - iog2.numberOfIObjects()
		     + count);
	    }
	}

	pairsSS /= 2;
	pairsSD /= 2;
	pairsDS /= 2;
	pairsDD /= 2;
	
	double randStatistic = (pairsSS+pairsDD)/totNumOfPairs;
	double jaccardCoefficient = pairsSS/(pairsSS+pairsSD+pairsDS);
	double folkesAndMallows = Math.sqrt(pairsSS*pairsSS /
				     (pairsSS + pairsSD) /
				     (pairsSS + pairsDS));
	*/

	double tp=0; // True positives
	double tn=0; // True negatives
	double fp=0; // False positives
	double fn=0; // False negatives

	Vector<IObjectGroup> g1groups = evalGrouping.getGroups();
	for(int i = 0; i < g1groups.size(); i++){
	    IObjectGroup iog1 = g1groups.get(i);
	    Vector<IObjectGroup> g2groups = refGrouping.getGroups();
	    for(int j = 0; j < g2groups.size(); j++){
		IObjectGroup iog2 = g2groups.get(j);
		double count = 
		    (Integer) confusionMatrix.get(iog1, iog2);

		tp += count * (count-1);
		fp += count * 
		    (iog2.numberOfIObjects() - count);
		fn += count * 
		    (iog1.numberOfIObjects() - count);
		tn += count * 
		    (numOfIOs 
		     - iog1.numberOfIObjects()
		     - iog2.numberOfIObjects()
		     + count);
	    }
	}

	tp /= 2;
	tn /= 2;
	fp /= 2;
	fn /= 2;
	
	double randStatistic = (tp+tn)/totNumOfPairs;
	double jaccardCoefficient = tp/(tp+fp+fn);
	double folkesAndMallows = Math.sqrt(tp * tp /
				     (tp + fp) /
				     (tp + fn));
	
	double precision = tp/(tp+fp);
	double recall = tp/(tp+fn);
	double beta = 0.5;
	double fmeasure = 
	    ((Math.pow(beta, 2) + 1)*precision*recall)/
	    (Math.pow(beta, 2)*precision + recall); 
	    

	
	MeasureGroup mg = new MeasureGroup("Global Extrinsic Pair Measures");
		
	Measure m = new Measure("Rand Statistic", randStatistic, 
				"Rand Statistic");
	mg.addMeasure(m);

	m = new Measure("Precision", precision, "Precision");
	mg.addMeasure(m);

	m = new Measure("Recall", recall, "Recall");
	mg.addMeasure(m);

	m = new Measure("F-Measure", fmeasure, "F-Measure (Beta=0.5)");
	mg.addMeasure(m);

	m = new Measure("Jaccard Coefficient", jaccardCoefficient, "Jaccard Coefficient");
	mg.addMeasure(m);

	m = new Measure("Folkes And Mallows", folkesAndMallows, "Folkes And Mallows");
	mg.addMeasure(m);

	m = new Measure("True Positives", tp, "True Positives");
	mg.addMeasure(m);

	m = new Measure("True Negatives", tn, "True Negatives");
	mg.addMeasure(m);

	m = new Measure("False Positives", fp, "False Positives");
	mg.addMeasure(m);

	m = new Measure("False Negatives", fn, "False Negatives");
	mg.addMeasure(m);


	return mg;
    }



	
    public MeasureGroup calculateGroupMeasures() throws MeasureNameConflictException{

	ExtrinsicGroupEvaluation extrinsicGroupEvaluation = 
	    new ExtrinsicGroupEvaluation(confusionMatrix, evalGrouping, refGrouping);

	MeasureGroup clusterGroup = 
	    new MeasureGroup("Extrinsic Group Measures");
	
	HashMap<IObjectGroup, Measure> groupPurityMap = extrinsicGroupEvaluation.getGroupPurityMap();

	HashMap<IObjectGroup, Measure> groupEntropyMap = extrinsicGroupEvaluation.getGroupEntropyMap();

	HashMap<IObjectGroup, Measure> groupNormEntropyMap = extrinsicGroupEvaluation.getGroupNormEntropyMap();

	Vector<IObjectGroup> g1groups = evalGrouping.getGroups();
	for(int i = 0; i < g1groups.size(); i++){
	    IObjectGroup iog1 = g1groups.get(i);
	    MeasureGroup oneGroupGroup = 
		calculateOneGroupMeasures(iog1, i, groupPurityMap, groupEntropyMap, groupNormEntropyMap);
	    clusterGroup.addMeasure(oneGroupGroup);
	}

	return clusterGroup;
    }

	
    private MeasureGroup calculateOneGroupMeasures(IObjectGroup iog, int groupNumber, HashMap<IObjectGroup, Measure> groupPurityMap, HashMap<IObjectGroup, Measure> groupEntropyMap, HashMap<IObjectGroup, Measure> groupNormEntropyMap) throws MeasureNameConflictException{
	MeasureGroup clusterGroup = new MeasureGroup(iog.getName());
	Measure m = new Measure("Group Size", iog.numberOfIObjects(), "Group Size");
	clusterGroup.addMeasure(m);

	m=groupPurityMap.get(iog);
	clusterGroup.addMeasure(m);

	m=groupEntropyMap.get(iog);
	clusterGroup.addMeasure(m);
	
	m=groupNormEntropyMap.get(iog);
	clusterGroup.addMeasure(m);
	
	return clusterGroup;
    }



    public double purity(){
	Vector<IObjectGroup> g1groups = evalGrouping.getGroups();

	int nmax = 0;
	for(int i = 0; i < g1groups.size(); i++){
	    IObjectGroup iog1 = g1groups.get(i);
	    Vector vec = confusionMatrix.valuesForFirstIndex(iog1);
	    if(vec==null)
		continue;
	    int count = (Integer) vec.get(0);
	    for(int j = 1; j < vec.size(); j++){
		if((Integer) vec.get(j)>count)
		    count = (Integer) vec.get(j);
	    }
	    nmax += count;
	}
	// purity:
	return (double) nmax / evalGrouping.numberOfIObjects();
    }


    public double entropy(){


	Vector<IObjectGroup> g1groups = evalGrouping.getGroups();

	double entropy = 0.0f;
	for(int i = 0; i < g1groups.size(); i++){
	    IObjectGroup iog1 = g1groups.get(i);
	    int iog1size = iog1.numberOfIObjects();

	    Vector g2groups = refGrouping.getGroups();
	    double iog1Entropy = 0.0;
	    for(int j = 0; j < g2groups.size(); j++){
		if((Integer) confusionMatrix.get(iog1, g2groups.get(j)) > 0){
		    double count = (Integer) confusionMatrix.get(iog1, g2groups.get(j));
		    double p = count/iog1.numberOfIObjects();
		    iog1Entropy -= p*MoreMath.log2(p);
		}
	    }
	    entropy += iog1size*iog1Entropy;
	}
	entropy /= evalGrouping.numberOfIObjects();

	return entropy;
    }


    /*
    public double entropy(){

	System.out.println("entropy");

	Vector<IObjectGroup> g1groups = evalGrouping.getGroups();

	double entropy = 0.0f;
	for(int i = 0; i < g1groups.size(); i++){
	    IObjectGroup iog1 = g1groups.get(i);
	    int iog1size = iog1.numberOfIObjects();
	    System.out.println("iog1: "+iog1+", "+iog1size);
	    double iog1entropy = 0.0f;
	    Vector vec = confusionMatrix.valuesForFirstIndex(iog1);
	    if(vec==null)
		continue;
	    for(int j = 0; j < vec.size(); j++){
		double p = ((double) ((Integer) vec.get(j)))/iog1size;
		if(p>0){
		    iog1entropy -= p*MoreMath.log2(p);
		    //		    System.out.println(" "+p);
		    System.out.println(" "+(-p*MoreMath.log2(p)));
		}
	    }
	    System.out.println(" iog1entropy "+iog1entropy);
	    entropy += iog1size * iog1entropy;
	    System.out.println(" entropy "+iog1size * iog1entropy);

	}
	System.out.println("entropy "+entropy);

	entropy /= evalGrouping.numberOfIObjects();

	System.out.println("div: "+evalGrouping.numberOfIObjects());
	System.out.println("entropy "+entropy);

	return entropy;
    }
    */

    public double mutualInformation(){
	double mutualInformation = 0;
	Vector<IObjectGroup> g1groups = evalGrouping.getGroups();
	Vector<IObjectGroup> g2groups = refGrouping.getGroups();
		
	double n = evalGrouping.numberOfIObjects();
	for(int i = 0; i < g1groups.size(); i++){
	    IObjectGroup iog1 = g1groups.get(i);
	    for(int j = 0; j < g2groups.size(); j++){
		IObjectGroup iog2 = g2groups.get(j);
		if(iog1.numberOfIObjects() > 0 && iog2.numberOfIObjects() > 0 && iog1.numberOfSharedIObjects(iog2) > 0){
		    double mij = iog1.numberOfSharedIObjects(iog2);
		    double ni = iog1.numberOfIObjects();
		    double nj = iog2.numberOfIObjects();
		    mutualInformation += (double) 
			mij*MoreMath.log2((mij*n)/(ni*nj));
		    //		    mutualInformation += (double) iog1.numberOfSharedIObjects(iog2)*MoreMath.log2(((double) iog1.numberOfSharedIObjects(iog2)*evalGrouping.numberOfIObjects())/(iog1.numberOfIObjects()*iog2.numberOfIObjects()));
		}
	    }
	}
	mutualInformation /= n;
	return mutualInformation;
    }
	
    public double normMutualInformation(){
	double evalEntropy = entropyForIObjectGroups(evalGrouping);
	double refEntropy = entropyForIObjectGroups(refGrouping);
	
	double mutualInformation = mutualInformation();
	
	return mutualInformation/Math.sqrt(evalEntropy * refEntropy);
    }

    public double groupsNormMutualInformation(){
	int evalGroups = evalGrouping.numberOfIObjectGroups();
	int refGroups = refGrouping.numberOfIObjectGroups();

	double mutualInformation = mutualInformation();
	
	return 2.0 * mutualInformation /
	    MoreMath.log2(evalGroups * refGroups);
    }


    /*
    public double normMutualInformation(){
	double normMutualInformation;
	double mutualInformation = 0;
	Vector<IObjectGroup> g1groups = evalGrouping.getGroups();
	Vector<IObjectGroup> g2groups = refGrouping.getGroups();
		
	double n = evalGrouping.numberOfIObjects();
	for(int i = 0; i < g1groups.size(); i++){
	    IObjectGroup iog1 = g1groups.get(i);
	    for(int j = 0; j < g2groups.size(); j++){
		IObjectGroup iog2 = g2groups.get(j);
		if(iog1.numberOfIObjects() > 0 && iog2.numberOfIObjects() > 0 && iog1.numberOfSharedIObjects(iog2) > 0){
		    double mij = iog1.numberOfSharedIObjects(iog2);
		    double ni = iog1.numberOfIObjects();
		    double nj = iog2.numberOfIObjects();
		    mutualInformation += (double) 
			mij*MoreMath.log2((mij*n)/(ni*nj));
		    //		    mutualInformation += iog1.numberOfSharedIObjects(iog2)*MoreMath.log2(((double) iog1.numberOfSharedIObjects(iog2)*evalGrouping.numberOfIObjects())/iog1.numberOfIObjects()*iog2.numberOfIObjects());
		}
	    }
	}
	mutualInformation /= n;
	normMutualInformation = (double) (mutualInformation*2.0/MoreMath.log2(evalGrouping.numberOfIObjectGroups()*refGrouping.numberOfIObjectGroups()));
	return normMutualInformation;
    }
    */

    /************************************/

    public Object getPropertyValue(String name){
    	return properties.getPropertyValue(name);
    }

    public void setProperty(String name, Object value) throws PropertyNotFoundException{
    	properties.setProperty(name, value);
    }

    public Properties getProperties(){
    	return properties;
    }

    public void setProperties(Properties properties){
    	this.properties=properties;
    }	
}
