package infomat.vectorspace.weighting;

import infomat.vectorspace.*;
import infomat.vectorspace.vector.*;
import infomat.vectorspace.similarity.*;

import mro.util.*;
import mro.util.properties.*;

import java.lang.Math.*;
import java.util.*;

/**
 * A class for weighting of a IMatrix. 
 * Weights the matrix using the current weighting and the
 * given ISimilarity and IObjectGrouping for the rows. 
 * Does nothing if the IObjectGrouping or ISimialrity are not set.
 *
 * @version          081119
 * @author           Magnus Rosell
 */
public class CentroidEntropyWeighter extends Weighter{

    private IObjectGrouping rowGrouping=null;
    private SparseISimilarity rowSimilarity=null;

    public CentroidEntropyWeighter(){
	super();
    }

    public CentroidEntropyWeighter(mro.util.properties.Properties properties){
	super(properties);
    }

    public mro.util.properties.Properties getDefaultProperties(){

	mro.util.properties.Properties properties = new mro.util.properties.Properties("Centroid Entropy Weighter Properties", CurrentClassGetter.getClassName());

	try{
	    Property p = new Property("Multiply with current weight", true, "Multiply with current weight");
	    PropertyValueDescription pvd = new PropertyValueDescription(true, "Do it.");
	    p.addValueDescription(pvd);
	    pvd = new PropertyValueDescription(false, "Do not do it.");
	    p.addValueDescription(pvd);
	    properties.addProperty(p);

	    p = new Property("Normalize rows", true, "Normalize rows");
	    pvd = new PropertyValueDescription(true, "Do it.");
	    p.addValueDescription(pvd);
	    pvd = new PropertyValueDescription(false, "Do not do it.");
	    p.addValueDescription(pvd);
	    properties.addProperty(p);

	}catch(PropertyNameConflictException pnce){
	    // Won't happen here
	}

	return properties;
    }

    public Object getPropertyValue(String name){
	return properties.getPropertyValue(name);
    }

    public void setProperty(String name, Object value) 
	throws PropertyNotFoundException{
	properties.setProperty(name, value);
    }

    public mro.util.properties.Properties getProperties(){
	return properties;
    }

    public void setProperties(mro.util.properties.Properties properties){
	this.properties=properties;
    }

    public String getWeighterName(){
	return "Centroid Entropy-Weighter";
    }

    
    public void setRowIObjectGrouping(IObjectGrouping rowGrouping){
	this.rowGrouping=rowGrouping;
    }

    public void setRowISimilarity(SparseISimilarity rowSimilarity){
	this.rowSimilarity=rowSimilarity;
    }


    /**
     * Weight the IMatrix (the whole matrix, 
     * which may be more than the current IObjectGroupings, the picture).
     * <br>
     * <br>
     * Does not work properly before #setRowIObjectGrouping()
     * and #setRowISimilarity() are called.
     */
    public void weight(IMatrix matrix){ 

	if(rowGrouping==null || rowSimilarity==null)
	    return;

	boolean useCurrentWeight  = 
	    (Boolean) getPropertyValue("Multiply with current weight");

	HashMap<IObjectGroup, IVector<IVectorIndex>> groupIVecMap = 
	    new HashMap<IObjectGroup, IVector<IVectorIndex>>();
	Vector<IObjectGroup> groups = rowGrouping.getGroups();
	for(int j=0; j<groups.size(); j++){
	    IObjectGroup g = groups.get(j);
	    groupIVecMap.put(g, rowSimilarity.getGroupRepresentation(g));
	}

	double maxEntropy = MoreMath.log(groups.size(),2);

	Vector<IObject> ios = matrix.getColumnIObjectSet().getOrdered();
	for(int i=0; i<ios.size(); i++){
	    IObject io = ios.get(i);

	    double[] pseudoProb = new double[groups.size()];
	    double pseudoProbSum = 0.0;
	    for(int j=0; j<groups.size(); j++){
		IVector<IVectorIndex> ivec = groupIVecMap.get(groups.get(j));
		Float val = ivec.getValue(io);
		if(val!=null){
		    pseudoProb[j] = val;
		    pseudoProbSum += val;
		}

	    }

	    double entropy = 0.0;
	    for(int j=0; j<groups.size(); j++){
		if(pseudoProb[j]>0.0){
		    double p = pseudoProb[j]/pseudoProbSum;
		    entropy -= p * MoreMath.log(p, 2);
		}
	    }	    

	    entropy = 1 - entropy/maxEntropy;
	    
	    matrix.setColumnIObjectGlobalWeight(io, (float) entropy);

	    //	    System.out.print(io.getString()+": "+entropy+", ");
	}
	//	System.out.println();

	Vector cellVec = matrix.getAllCells();
	for(int i=0; i<cellVec.size(); i++){
	    IMatrixCell mc = (IMatrixCell) cellVec.get(i);
	    float w = 
		matrix.getColumnIObjectGlobalWeight(mc.getColumnIObject());
	    if(useCurrentWeight)
		mc.setWeight(mc.getWeight() * w);
	    else 
		mc.setWeight(w);
	}

	boolean normalize  = 
	    (Boolean) getPropertyValue("Normalize rows");
	if(normalize)
	    Normalizer.normalizeMatrix(matrix, true);

    }
}
