package mro.util.doublehashmap;

import java.util.*; 

/**
 * A DoubleHashMap has two indexes. 
 * There is only one copy of each index value.
 *
 * @version          080429
 * @author           Magnus Rosell
 */
public class GenericDoubleHashMap<F, S, V>{

    protected HashSet<F> firstIndexSet = new HashSet<F>();
    protected HashSet<S> secondIndexSet = new HashSet<S>();
	
    // The Map, i.e. the first HashMap, which contains several HashMaps,
    // HashMap:s for the second index.
    protected HashMap<F, HashMap<S, V>> firstMap = 
	new HashMap<F, HashMap<S, V>>();

    // The alternate view.
    // The second HashMap, which contains several HashMaps,
    // HashMap:s for the first index.
    protected HashMap<S, HashMap<F, V>> secondMap = 
	new HashMap<S, HashMap<F, V>>();

    // The number of Objects in the DoubleHashMap
    protected int size=0;

    private Class<F> fClass=null;
    private Class<S> sClass=null;
    private Class<V> vClass=null;

    /**
     * Constructor.
     */
    public GenericDoubleHashMap(){}

    /**
     * Constructor. Use this when the first and second index 
     * has the same class as in these cases the invert()-method 
     * becomes faster.
     */
    public GenericDoubleHashMap(Class<F> fClass, Class<S> sClass, 
			      Class<V> vClass){ 
	this.fClass=fClass;
	this.sClass=sClass;
	this.vClass=vClass;
    }

    public GenericDoubleHashMap(GenericDoubleHashMap<F, S, V> map){
	firstIndexSet = new HashSet<F>(map.firstIndexSet);
	secondIndexSet = new HashSet<S>(map.secondIndexSet);

	firstMap = new HashMap<F, HashMap<S, V>>(map.firstMap);
	secondMap = new HashMap<S, HashMap<F, V>>(map.secondMap);
	
	this.size = map.size;

	this.fClass=map.fClass;
	this.sClass=map.sClass;
	this.vClass=map.vClass;
    }


    /**
     * Adds a Object "value" indexed with the Object:s
     * "first" and "second"
     * @param first
     * @param second
     * @param value
     */
    public void put(F first, S second, V value){
	firstIndexSet.add(first);
	secondIndexSet.add(second);
		
	putIntoFirstMap(first, second, value, firstMap);
	putIntoSecondMap(second, first, value, secondMap);

	size++;
    }

    /**
     * Convenience method, for putting objects into a HashMap of HashMap:s.
     */
    protected void putIntoFirstMap(F o1, S o2, V value, HashMap<F, HashMap<S, V>> map){
	HashMap<S, V> map2 = map.get(o1);
	if(map2==null){
	    map2 = new HashMap<S, V>();
	    map.put(o1, map2);
	}
	map2.put(o2, value);
    }

    /**
     * Convenience method, for putting objects into a HashMap of HashMap:s.
     */
    protected void putIntoSecondMap(S o1, F o2, V value, HashMap<S, HashMap<F, V>> map){
	HashMap<F, V> map2 = map.get(o1);
	if(map2==null){
	    map2 = new HashMap<F, V>();
	    map.put(o1, map2);
	}
	map2.put(o2, value);
    }


    /**
     * Returns the Object that was indexed with "first" and "second".
     * @param first
     * @param second
     * @return an Object
     */
    public V get(F first, S second){
	HashMap<S, V> secondMap = firstMap.get(first);
	if(secondMap==null)
	    return null;
	V value = secondMap.get(second);
	if(value==null)
	    return null;
	return value;
    }

    /**
     * Removes the Object with the indexes.
     * @param first
     * @param second
     * @return an Object
     */
    public V remove(F first, S second){

	// The following two must be the same.
	// No check right now. It should not be possible
	// for it to go wrong.
	V value = removeFromFirstMap(first, second, firstMap);
	value = removeFromSecondMap(second, first, secondMap);

	if(firstMap.get(first)==null)
	    firstIndexSet.remove(first);
	
	if(secondMap.get(second)==null)
	    secondIndexSet.remove(second);

	size--;
	return value;
    }

    /**
     * Convenience method, for removing objects in a HashMap of HashMap:s.
     */
    private V removeFromFirstMap(F o1, S o2, HashMap<F, HashMap<S, V>> map){
	HashMap<S, V> map2 = map.get(o1);
	if(map2==null)
	    return null;
	V value = map2.remove(o2);
	if(value==null)
	    return null;
	if(map2.size()==0)
	    map.remove(map2);
	return value;
    }

    /**
     * Convenience method, for removing objects in a HashMap of HashMap:s.
     */
    private V removeFromSecondMap(S o1, F o2, HashMap<S, HashMap<F, V>> map){
	HashMap<F, V> map2 = map.get(o1);
	if(map2==null)
	    return null;
	V value = map2.remove(o2);
	if(value==null)
	    return null;
	if(map2.size()==0)
	    map.remove(map2);
	return value;
    }

    /**
     * Removes all Objects with the specified "first" index.
     * @param first
     */
    public void removeAllFirst(F first){
	HashMap<S, V> removed = firstMap.remove(first);
	size-=removed.size();

	Iterator<HashMap<F, V>> it = secondMap.values().iterator();
	while(it.hasNext()){
	    HashMap<F, V> tempFirstMap = it.next();
	    tempFirstMap.remove(first);
	}

	firstIndexSet.remove(first);
    }

    /**
     * Removes all Objects with the specified "second" index.
     * @param second
     */
    public void removeAllSecond(S second){
	HashMap<F, V> removed = secondMap.remove(second);
	size-=removed.size();

	Iterator<HashMap<S, V>> it = firstMap.values().iterator();
	while(it.hasNext()){
	    HashMap<S, V> tempSecondMap = it.next();
	    tempSecondMap.remove(second);
	}

	secondIndexSet.remove(second);
    }


    /**
     * Returns all Objects in the DoubleHashMap as a Vector.
     * @return a Vector
     */
    public Vector<V> values(){
	Vector<V> ret = new Vector<V>();
	Iterator<HashMap<S, V>> it = firstMap.values().iterator();
	while(it.hasNext()){
	    HashMap<S, V> secondMap = it.next();
	    Iterator<V> it2 = secondMap.values().iterator();
	    while(it2.hasNext())
		ret.add(it2.next());
	}
	return ret;
    }

    /**
     * Returns a Vector of those value Objects that are 
     * stored with the specified "first" index.
     * @param first
     * @return a Vector
     */	
    public Vector<V> valuesForFirstIndex(F first){
	HashMap<S, V> second = firstMap.get(first);
	if(second==null)
	    return null;
	return new Vector<V>(second.values());
    }

    /**
     * Returns a Vector of those value Objects that are 
     * stored with the specified "second" index.
     * @param second
     * @return a Vector
     */	
    public Vector<V> valuesForSecondIndex(S second){
	HashMap<F, V> first = secondMap.get(second);
	if(first==null)
	    return null;
	return new Vector<V>(first.values());
    }

    /**
     * Inverts the mapping, so the fist index becomes the second.
     * Makes a copy!
     */
    public GenericDoubleHashMap<S, F, V> invert(){

	GenericDoubleHashMap<S, F, V> newMap = 
	    new GenericDoubleHashMap<S, F, V>(sClass, fClass, vClass);

	if(fClass!=null && fClass == sClass){
	    newMap.firstIndexSet = new HashSet<S>(secondIndexSet);
	    newMap.secondIndexSet = new HashSet<F>(firstIndexSet);
	    
	    newMap.firstMap = new HashMap<S, HashMap<F, V>>(secondMap);
	    newMap.secondMap = new HashMap<F, HashMap<S, V>>(firstMap);

	    newMap.size = size;
	}else{
	    Iterator<Map.Entry<F, HashMap<S, V>>> it = 
		firstMap.entrySet().iterator();
	    while(it.hasNext()){
		Map.Entry<F, HashMap<S, V>> entry = it.next();
		F f = entry.getKey();
		HashMap<S, V> map = entry.getValue();
		Iterator<Map.Entry<S, V>> it2 = map.entrySet().iterator();
		while(it2.hasNext()){
		    Map.Entry<S, V> entry2 = it2.next();
		    S s = entry2.getKey();
		    V v = entry2.getValue();
		    newMap.put(s, f, v);
		}
	    }
	}
	
	return newMap;
    }

    public int size(){
	return size;
    }

    public int sizeOfFirstIndex(){
	return firstIndexSet.size();
    }

    public int sizeOfSecondIndex(){
	return secondIndexSet.size();
    }

    public int sizeOfSecondIndexForElementInFirst(F first){
	HashMap<S, V> second = firstMap.get(first);
	if(second==null)
	    return 0;
	return second.size();
    }

    public Set<S> keySetOfSecondIndexForElementInFirst(F first){
	HashMap<S, V> second = firstMap.get(first);
	if(second==null)
	    return null;
	return new HashSet<S>(second.keySet());
    }

}
