/*
 *  Copyright (C) 2006 Derek Greene
 *
 *  This file is part of TCT.
 *
 *  TCT is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  TCT is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with TCT; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
/**
 * C code for JNI implementation of actual factorization stage of the
 * Lee-Seung Non-negative Matrix Factorization (NMF) approach that minimizes
 * the Kullback-Leibler divergence (i.e. relative entropy).
 */

#include <stdio.h>
#include <math.h>
#include <string.h>
#include "NativeKL.h"
#include "Matrix.h"

// If defined, demand Java to allow access to original arrays not copies
#define USE_CRITICAL 1
// Number of iterations at which to check convergence
#define CONVERGENCE_CHECK_COUNT 5

static const double EPSILON = 1e-15;

int factorDense( const int k, const int m, const int n, double *dataA, double *dataU, double *dataV, const int minIterations, const int maxIterations, const double threshold );

int factorSparse( SparseRowMatrix *A, const int k, const int m, const int n, double *dataA, int *rowPointer, int *used, int *columnIndex, double *dataU, double *dataV, const int minIteration, const int maxIterations, const double threshold );

bool isConverged( double *dataX, double *dataY, const int rows, const int cols, const double threshold );

// --------------------------------------------------------------------------------------------------------
// JNI METHODS
// --------------------------------------------------------------------------------------------------------

/*
 * Class:     nmf_nni_NativeKL
 * Method:    factorDense
 * Signature: (II[D[D[DI)I
 */
JNIEXPORT jint JNICALL Java_core_clustering_nmf_jni_NativeKLFactorization_factorDense
  (JNIEnv *env, jobject self,
	jint jK, jint jRowsA, jint jColsA, jdoubleArray jA, jdoubleArray jU,jdoubleArray jV, 
	jint jMinIterations, jint jMaxIterations, jdouble jThreshold )
{
	double *a, *u, *v;
	jboolean aIsCopy, uIsCopy, vIsCopy;
	int iteration;

	/// Access arrays
#ifdef USE_CRITICAL
	a = (double *)env->GetPrimitiveArrayCritical( jA, &aIsCopy );
	u = (double *)env->GetPrimitiveArrayCritical( jU, &uIsCopy );
	v = (double *)env->GetPrimitiveArrayCritical( jV, &vIsCopy );
#else
	a = env->GetDoubleArrayElements( jA, &aIsCopy );
	u = env->GetDoubleArrayElements( jU, &uIsCopy );
	v = env->GetDoubleArrayElements( jV, &vIsCopy );
#endif

	/// Factor
	iteration = factorDense( jK, jRowsA, jColsA, a, u, v, jMinIterations, jMaxIterations, jThreshold );

	/// Free memory
#ifdef USE_CRITICAL
	env->ReleasePrimitiveArrayCritical( jA, a, JNI_ABORT );
	env->ReleasePrimitiveArrayCritical( jU, u, JNI_FALSE );
	env->ReleasePrimitiveArrayCritical( jV, v, JNI_FALSE );
#else
	env->ReleaseDoubleArrayElements( jA, a, JNI_ABORT );
	env->ReleaseDoubleArrayElements( jU, u, JNI_FALSE );
	env->ReleaseDoubleArrayElements( jV, v, JNI_FALSE );
#endif

	return iteration;
}

 
/*
 * Class:     nmf_nni_NativeKL
 * Method:    factorSparse
 * Signature: (III[D[I[I[I[D[DIID)I
 */
JNIEXPORT jint JNICALL Java_core_clustering_nmf_jni_NativeKLFactorization_factorSparse
  (JNIEnv *env, jobject self,
	jint jK, jint jRowsA, jint jColsA, jdoubleArray jA, jintArray jRowPointer, jintArray jUsed, jintArray jColumnIndex, jdoubleArray jU, jdoubleArray jV, 
	jint jMinIterations, jint jMaxIterations, jdouble jThreshold )
{
	double *a, *u, *v;
	jboolean aIsCopy, uIsCopy, vIsCopy;
	int *rowPointer, *used, *columnIndex;
	jboolean dataIsCopy, rowPointerIsCopy, usedIsCopy, colIndexIsCopy;
	int iteration;

	/// Access arrays
#ifdef USE_CRITICAL
	a = (double *)env->GetPrimitiveArrayCritical( jA, &aIsCopy );
	u = (double *)env->GetPrimitiveArrayCritical( jU, &uIsCopy );
	v = (double *)env->GetPrimitiveArrayCritical( jV, &vIsCopy );
	// Handle matrix specific items
	rowPointer = (int *)env->GetPrimitiveArrayCritical( jRowPointer, &rowPointerIsCopy );
	columnIndex = (int *)env->GetPrimitiveArrayCritical( jColumnIndex, &colIndexIsCopy );
	used = (int *)env->GetPrimitiveArrayCritical( jUsed, &usedIsCopy );
#else
	a = env->GetDoubleArrayElements( jA, &aIsCopy );
	u = env->GetDoubleArrayElements( jU, &uIsCopy );
	v = env->GetDoubleArrayElements( jV, &vIsCopy );
	// Handle matrix specific items
	rowPointer = env->GetIntArrayElements( jRowPointer, &rowPointerIsCopy );
	used = env->GetIntArrayElements( jUsed, &usedIsCopy );
	columnIndex = env->GetIntArrayElements( jColumnIndex, &colIndexIsCopy );
#endif

	/// Create matrix
	SparseRowMatrix *A = new SparseRowMatrix( jRowsA, jColsA, a, rowPointer, used, columnIndex );
	/// Factor
	iteration = factorSparse( A, jK, jRowsA, jColsA, a, rowPointer, used, columnIndex, u, v, jMinIterations, jMaxIterations, jThreshold );

	/// Free memory
	delete A;
#ifdef USE_CRITICAL
	env->ReleasePrimitiveArrayCritical( jA, a, JNI_ABORT );
	env->ReleasePrimitiveArrayCritical( jU, u, JNI_FALSE );
	env->ReleasePrimitiveArrayCritical( jV, v, JNI_FALSE );
	// NB: The JNI_ABORT flag specifies that if you are using a local copy of the array, there is no need to copy back to the real array.
	env->ReleasePrimitiveArrayCritical( jRowPointer, rowPointer, JNI_ABORT );
	env->ReleasePrimitiveArrayCritical( jUsed, used, JNI_ABORT );
	env->ReleasePrimitiveArrayCritical( jColumnIndex, columnIndex, JNI_ABORT );
#else
	env->ReleaseDoubleArrayElements( jA, a, JNI_ABORT );
	env->ReleaseDoubleArrayElements( jU, u, JNI_FALSE );
	env->ReleaseDoubleArrayElements( jV, v, JNI_FALSE );
	env->ReleaseIntArrayElements( jRowPointer, rowPointer, JNI_ABORT );
	env->ReleaseIntArrayElements( jUsed, used, JNI_ABORT );
	env->ReleaseIntArrayElements( jColumnIndex, columnIndex, JNI_ABORT );
#endif

	return iteration;
}

// --------------------------------------------------------------------------------------------------------
// NMF METHODS
// --------------------------------------------------------------------------------------------------------

int factorDense( const int k, const int m, const int n, double *dataA, double *dataU, double *dataV, const int minIterations, const int maxIterations, const double threshold )
{
	const size_t lengthU = m*k;
	const size_t lengthV = n*k;

	// Tempory storage
	double *lastU = new double[lengthU];
	double *lastV = new double[lengthV];
	// Temporary variables
	double uv1, uv2, vt, ut, norm;
	int iteration, x, y, row, col;
	
	double a, change;
	for (iteration = 1; iteration <= maxIterations; iteration++)
	{
		/// Update V based on KL update rule
		// void *memcpy(void *dest, const void *src, size_t n);
		memcpy( lastV, dataV, lengthV * sizeof(double) );
		for (row = 0; row < n; row++)
		{
			for (col = 0; col < k; col++)
			{
				//	TransAMultB: VT <- (UV)' * U
				vt = 0;
				for ( x = 0; x < m; x++ )
				{
					a = dataA[ x + row * m];
					// don't bother if a=0
					if( a != 0 )
					{
						// MultATransB: UV <- U*V' : Get(x,row)
						uv1 = 0;
						for ( y = 0; y < k; y++ )
						{
							uv1 += dataU[x + y * m] * lastV[row + y * n];
						}
						//	SaveInvDiv: UV <- A ./ (UV+eps) : Set A(x,row) / UV(x.row) + EPS
						uv2 = a / ( uv1 + EPSILON );
						vt += uv2 * dataU[x + col * m];
					}
				}
				// Dot Multiplication: V <- V .* VT : Set(row, col)
				if( vt != 0 )
				{
					dataV[row + col * n] = lastV[row + col * n] * vt;
				}
			}
		}

		/// Update U based on KL update rule
		// void *memcpy(void *dest, const void *src, size_t n);
		memcpy( lastU, dataU, lengthU * sizeof(double) );
		for (row = 0; row < m; row++)
		{
			for (col = 0; col < k; col++)
			{
				//	Mult: UT <- UV * V : Get(row, col)
				ut = 0;
				for ( x = 0; x < n; x++ )
				{
					// MultATransB: UV <- U*V' : Get(row, x)
					uv1 = 0;
					for ( y = 0; y < k; y++ )
					{
						uv1 += lastU[row + y * m] * dataV[x + y * n];
					}
					//	SaveInvDiv: UV <- A ./ (UV+eps) : Set A(row,x) / UV(row, x) + EPS
					uv2 = dataA[row + x*m] / ( uv1 + EPSILON );
					ut += uv2 * dataV[x + col * n];
				}
				// Dot Multiplication:  U <- U .* UT : Set(row,col)
				dataU[row + col * m ] = lastU[row + col * m] * ut;
			}
		}
			
		// Apply Fast L1 column normalisation of U
		for (col = 0; col < k; col++)
		{
			// calculate norm
			norm = 0;
			for( row = 0; row < m; row++ )
			{
				norm += dataU[row + col * m ];
			}
			// normalise this column
			if( norm != 0.0 )
			{
				for( row = 0; row < m; row++ )
				{
					dataU[row + col * m ] /= norm;
				}
			}
		}
					
		/// Test for convergence
		if( (iteration %CONVERGENCE_CHECK_COUNT == 0) && isConverged( dataV, lastV, n, k, threshold ) && ( iteration >= minIterations ) )
		{
			break;
		}
	}

	delete[] lastU;
	delete[] lastV;

	return iteration;
}

int factorSparse( SparseRowMatrix *A, const int k, const int m, const int n, double *dataA, int *rowPointer, int *used, int *columnIndex, double *dataU, double *dataV, const int minIterations, const int maxIterations, const double threshold )
{
	const size_t lengthU = m*k;
	const size_t lengthV = n*k;

	// Tempory storage
	double *lastU = new double[lengthU];
	double *lastV = new double[lengthV];
	// Temporary variables
	double uv1, uv2, vt, ut, norm;
	int iteration, x, y, row, col;
	int low, valPos;
	
	double change;

	// Create dense matrix too
	double *denseData = new double[m * n];
	for (row = 0; row < m; row++)
	{
		for (col = 0; col < n; col++)
		{
			denseData[row + col * m] = A->get( row, col );
		}
	}

	/// Iterate until max iterations or converged 
	for (iteration = 1; iteration <= maxIterations; iteration++)
	{
		/// Update V based on KL update rule
		memcpy( lastV, dataV, lengthV * sizeof(double) );
		for (row = 0; row < n; row++)
		{
			for (col = 0; col < k; col++)
			{
				//	TransAMultB: VT <- (UV)' * U
				vt = 0;
				for ( x = 0; x < m; x++ )
				{
					// The access to A is expensive, so try to avoid it...
					if( dataU[x + col * m] != 0 )
					{
						// MultATransB: UV <- U*V' : Get(x,row)
						uv1 = 0;
						for ( y = 0; y < k; y++ )
						{
							uv1 += dataU[x + y * m] * lastV[row + y * n];
						}
						//	SaveInvDiv: UV <- A ./ (UV+eps) : Set A(x,row) / UV(x.row) + EPS
						uv2 = denseData[x + row * m] / ( uv1 + EPSILON );
						vt += uv2 * dataU[x + col * m];
					}
				}
				// Dot Multiplication: V <- V .* VT : Set(row, col)
				dataV[row + col * n] = lastV[row + col * n] * vt;
			}
		}

		/// Update U based on KL update rule
		memcpy( lastU, dataU, lengthU * sizeof(double) );
		for (row = 0; row < m; row++)
		{
			for (col = 0; col < k; col++)
			{
				//	Mult: UT <- UV * V : Get(row, col)
				ut = 0;
				// NB: we do this efficiently so that we don't need A.getQuick
				// and we don't do further processing on any zero values in A
				low = rowPointer[row];
				for (valPos = rowPointer[row + 1]; --valPos >= low;)
				{
					if( dataA[valPos] != 0 )
					{
						// NB: x = column index for this value
						x = columnIndex[valPos];
						uv1 = 0;
						for ( y = 0; y < k; y++ )
						{
							uv1 += lastU[row + y * m] * dataV[x + y * n];
						}
						uv2 = dataA[valPos] / ( uv1 + EPSILON );
						ut += uv2 * dataV[x + col * n];
					}
				}
				// Dot Multiplication:  U <- U .* UT : Set(row,col)
				dataU[row + col * m ] = lastU[row + col * m] * ut;
			}
		}
			
		// Apply Fast L1 column normalisation of U
		for (col = 0; col < k; col++)
		{
			// calculate norm
			norm = 0;
			for( row = 0; row < m; row++ )
			{
				norm += dataU[row + col * m ];
			}
			// normalise this column
			if( norm != 0.0 )
			{
				for( row = 0; row < m; row++ )
				{
					dataU[row + col * m ] /= norm;
				}
			}
		}

		/// Test for convergence
		if( (iteration % CONVERGENCE_CHECK_COUNT == 0) && isConverged( dataV, lastV, n, k, threshold ) && ( iteration >= minIterations ) )
		{
			break;
		}
	}
	
	delete[] denseData;
	delete[] lastU;
	delete[] lastV;

	return iteration;
}

// --------------------------------------------------------------------------------------------------------
// CONVERGENCE METHODS
// --------------------------------------------------------------------------------------------------------

/**
 * Determines factorisation convergence based size of changes
 * (as determined by Frobenius norm) in factors between iterations.
 */
bool isConverged( double *dataX, double *dataY, const int rows, const int cols, const double threshold )
{
	double fnorm = 0.0, d;
	int r, c;
	
	// Split in 3
	double sum = 0;
	for( r = 0; r < rows; r++ )
	{
		for( c = 0; c < cols; c++ )
		{
			// Calculate ( X.getQuick(r,c) - lastX.getQuick(r,c) );
			d = ( dataX[r + c * rows] - dataY[r + c * rows] );
			fnorm += (d * d);
		}
		// Early exit?
		if( r % 100 == 0 )
		{
			if( (sqrt(fnorm)/rows) > threshold )
			{
				return false;
			}
		}
	}
	fnorm = sqrt(fnorm) / rows;
	return ( fnorm <= threshold );
}
