/*
 * Copyright (C) 2006 Derek Greene
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <stdio.h>
#include "Arpack.h"
#include "ArpackEVD.h"
#include "Matrix.h"

// If defined, demand Java to allow access to original arrays not copies
#define USE_CRITICAL 1
#define DEBUG 0

int factorEVD( Matrix *mm, int nev, bool largest, bool computeVectors, double *evdS, double *evdE );
int getENCV( int n, int nev );
void populateEVD( int nev, double *dr, double *evdS, bool reverse );
void populateEVD( Matrix *mm, int nev, double *dr, double *v, double *evdS, DenseMatrix *E, bool reverse );

// --------------------------------------------------------------------------------------------------------
// JNI METHODS
// --------------------------------------------------------------------------------------------------------

/*
 * Class:     matrix_jni_arpack_ArpackEVD
 * Method:    factorDense
 * Signature: (IZI[D[D[D)I
 */
JNIEXPORT jint JNICALL Java_matrix_decomp_jni_ArpackEVD_factorDense
  (JNIEnv *env, jobject self, 
	jint jK, jboolean jLargest, jint jN, jdoubleArray jS, jdoubleArray jE, jdoubleArray jData )
{
	int error;
	double *s, *e, *data;
	jboolean eIsCopy, sIsCopy, dataIsCopy;
	
	/// Access arrays
#ifdef USE_CRITICAL
	// Handle EVD items
	s = (double *)env->GetPrimitiveArrayCritical( jS, &sIsCopy );
	e = (double *)env->GetPrimitiveArrayCritical( jE, &eIsCopy );
	// Handle matrix specific items
	data = (double *)env->GetPrimitiveArrayCritical( jData, &dataIsCopy );
#else
	// Handle EVD items
	s = env->GetDoubleArrayElements( jS, &sIsCopy );
	e = env->GetDoubleArrayElements( jE, &eIsCopy );
	// Handle matrix specific items
	data = env->GetDoubleArrayElements( jData, &dataIsCopy );
#endif

	if( DEBUG ) printf( "Critical=%d sIsCopy=%d eIsCopy=%d dataIsCopy=%d\n", USE_CRITICAL, sIsCopy, eIsCopy, dataIsCopy );

	/// Create matrix
	Matrix *mm = new DenseMatrix( jN, jN, data );

	/// Factor
	error = factorEVD( mm, jK, jLargest, true, s, e );

	/// Return the memory to Java
	delete mm;
#ifdef USE_CRITICAL
	env->ReleasePrimitiveArrayCritical( jS, s, JNI_FALSE );
	env->ReleasePrimitiveArrayCritical( jE, e, JNI_FALSE );
	env->ReleasePrimitiveArrayCritical( jData, data, JNI_ABORT );
#else
	env->ReleaseDoubleArrayElements( jS, s, JNI_FALSE );
	env->ReleaseDoubleArrayElements( jE, e, JNI_FALSE );
	env->ReleaseDoubleArrayElements( jData, data, JNI_ABORT );
#endif
	return error;
}

/*
 * Class:     matrix_jni_arpack_ArpackEVD
 * Method:    factorSparseRow
 * Signature: (IZI[D[D[D[I[I[I)I
 */
JNIEXPORT jint JNICALL Java_matrix_decomp_jni_ArpackEVD_factorSparseRow
  (JNIEnv *env, jobject self, 
	jint jK, jboolean jLargest, jint jN, jdoubleArray jS, jdoubleArray jE, jdoubleArray jData, jintArray jRowPointer, jintArray jUsed, jintArray jColumnIndex)
{
	int error;
	double *s, *e, *data;
	jboolean eIsCopy, sIsCopy;
	int *rowPointer, *used, *columnIndex;
	jboolean dataIsCopy, rowPointerIsCopy, usedIsCopy, colIndexIsCopy;
	
	/// Access arrays
#ifdef USE_CRITICAL
	// Handle EVD items
	s = (double *)env->GetPrimitiveArrayCritical( jS, &sIsCopy );
	e = (double *)env->GetPrimitiveArrayCritical( jE, &eIsCopy );
	// Handle matrix specific items
	data = (double *)env->GetPrimitiveArrayCritical( jData, &dataIsCopy );
	rowPointer = (int *)env->GetPrimitiveArrayCritical( jRowPointer, &rowPointerIsCopy );
	columnIndex = (int *)env->GetPrimitiveArrayCritical( jColumnIndex, &colIndexIsCopy );
	used = (int *)env->GetPrimitiveArrayCritical( jUsed, &usedIsCopy );
#else
	// Handle EVD items
	s = env->GetDoubleArrayElements( jS, &sIsCopy );
	e = env->GetDoubleArrayElements( jE, &eIsCopy );
	// Handle matrix specific items
	data = env->GetDoubleArrayElements( jData, &dataIsCopy );
	rowPointer = env->GetIntArrayElements( jRowPointer, &rowPointerIsCopy );
	used = env->GetIntArrayElements( jUsed, &usedIsCopy );
	columnIndex = env->GetIntArrayElements( jColumnIndex, &colIndexIsCopy );
#endif
	if( DEBUG ) printf( "Critical=%d sIsCopy=%d eIsCopy=%d dataIsCopy=%d\n", USE_CRITICAL, sIsCopy, eIsCopy, dataIsCopy );
	if( DEBUG ) printf( "rowPointerIsCopy=%d usedIsCopy=%d colIndexIsCopy=%d\n", rowPointerIsCopy, usedIsCopy, colIndexIsCopy );

	/// Create matrix
	Matrix *mm = new SparseRowMatrix( jN, jN, data, rowPointer, used, columnIndex );

	/// Factor
	error = factorEVD( mm, jK, jLargest, true, s, e );
	
	/// Return the memory to Java
	delete mm;
#ifdef USE_CRITICAL
	env->ReleasePrimitiveArrayCritical( jS, s, JNI_FALSE );
	env->ReleasePrimitiveArrayCritical( jE, e, JNI_FALSE );
	// NB: The JNI_ABORT flag specifies that if you are using a local copy of the array, there is no need to copy back to the real array.
	env->ReleasePrimitiveArrayCritical( jData, data, JNI_ABORT );
	env->ReleasePrimitiveArrayCritical( jRowPointer, rowPointer, JNI_ABORT );
	env->ReleasePrimitiveArrayCritical( jUsed, used, JNI_ABORT );
	env->ReleasePrimitiveArrayCritical( jColumnIndex, columnIndex, JNI_ABORT );
#else
	env->ReleaseDoubleArrayElements( jS, s, JNI_FALSE );
	env->ReleaseDoubleArrayElements( jE, e, JNI_FALSE );
	env->ReleaseDoubleArrayElements( jData, data, JNI_ABORT );
	env->ReleaseIntArrayElements( jRowPointer, rowPointer, JNI_ABORT );
	env->ReleaseIntArrayElements( jUsed, used, JNI_ABORT );
	env->ReleaseIntArrayElements( jColumnIndex, columnIndex, JNI_ABORT );
#endif
	return error;
}

// --------------------------------------------------------------------------------------------------------
// INTERNALS
// --------------------------------------------------------------------------------------------------------


int factorEVD( Matrix *mm, int nev, bool computeLargest, bool computeVectors, double *evdS, double *evdE )
{
	/// Temp variables
	int i, j;

	// Get matrix size
	int rows = mm->rows();
	int cols = mm->columns();
	int n = ( rows > cols ) ? rows : cols;
	
	/// ALLOCATE SPACE
	// Calculate ncv: The largenumber of basis vectors that will be used in the process.
	int ncv = getENCV( n, nev );
	int ldv = n;
	// Space for eigenvectors
	double *v = new double[ldv * ncv];
	// Workspace size
	int lworkl = ncv * (ncv + 8);
	// Space for workspace 
	double *workl = new double[lworkl];
	double *workd = new double[3 * n];
	// Space for eigenvalues (real and imaginary)
	double *d = new double[2 * ncv];
	// Space for residuals
	double *resid = new double[n];

	/// USER SETTINGS
	// Which eigenvalues to calculate?	
	char* sWhich;
	if( computeLargest )
	{
		// largest algebraic part...
		sWhich = "LA";
	}
	else
	{
		// smallest algebraic part...
		sWhich = "SA";
	}		
	// Specifies whether eigenvectors should be calculated 
	int rvec;		
	if( computeVectors )
	{
		rvec = 1;
	}
	else
	{
		rvec = 0;
	}

	/// INTERNAL SETTINGS
	char bmat[2] = "I";
	// Sets the tolerance
	double tol = 0.0;
	// An array used to pass information to the routines about their functional modes.
	int *iparam = new int[11];
	// Use the exact shift strategy
	iparam[0] = 1;
	// Set  max number of Arnoldi iterations
	iparam[2] = 1000;	
	// Set Mode 1 calculation - standard eigensystem
	iparam[6] = 1; 

	/// STATE
	int ido = 0;
	// Passes convergence information out of the iteration routine.
	int info = 0;
	// Returns info from dneupd
	int ierr;
	// Indicates the locations in the work array workd where the input and output vectors in the
	// callback routine are located
	int *ipntr = new int[14];	
	// TODO: what are these for?
	int *select = new int[ncv];
	double sigma;

	/// PHASE #1: Arnoldi Iteration
	do
	{ 
		// Repeatedly call the routine DNAUPD and take  actions indicated by parameter IDO until
		// either convergence is indicated or maxitr has been exceeded. 
		dsaupd(&ido, bmat, &n, sWhich, &nev, &tol, resid,
			&ncv, v, &ldv, iparam, ipntr, workd, workl,
			&lworkl, &info);
	  
		// Handle state
		if( ( ido == -1 ) || ( ido == 1 ) )
		{
			// find matrix-vector product
			mm->mult( workd + ipntr[0] - 1, workd + ipntr[1] - 1 );
		}
	}
	while( ( ido == 1 ) || ( ido == -1 ) );

	// printf( "Info=%d\n", info );
	// Any fatal errors?
	if( info < 0 )
	{
		//// printf( "Error: failure during decomposition, info=%d\n", info );
		return info;
	}
	// Any non-fatal errors
	else if( info == 1 )
	{
		//// printf( "Maximum number of iteration reached\n" );
		return info;
	}
	else if( info > 0 )
	{
		return info;
	}

	/// PHASE #2: Post-Processing

	// Extract the eigenvalues (and eigenvectors if rvec=1)
	dseupd(&rvec, "All", select, d, v, &ldv, &sigma, bmat,
		&n, sWhich, &nev, &tol, resid, &ncv, v, &ldv,
		iparam, ipntr, workd, workl, &lworkl, &ierr);			

	// Any fatal errors?
	if( ierr != 0 )
	{
		// printf( "Error: failure during eigenvalue extraction, info = %d\n", ierr );
		return ( ierr - 1000 );
	}
	
	// Populate EVD arrays
	if( computeVectors )
	{
		DenseMatrix *E = new DenseMatrix( rows, nev, evdE );
		populateEVD( mm, nev, d, v, evdS, E, true );
		delete E;
	}
	else
	{
		populateEVD( nev, d, evdS, true );
	}
	
	
	/// TIDY UP
	// deallocate workspace
	delete workl;
	delete workd;
	delete d;
	delete v;
	delete resid;
	// deallocate settings
	delete iparam;
	// deallocate state
	delete ipntr;
	delete select;

	// Success
	return 0;	
}


/**
 * Calculate the largest number of basis vectors that will be used in the Implicitly Restarted 
 * Arnoldi Process for symmetric and non-symmetric Eigenvalue decomposition.
 *
 * @param n      number of rows for eigenvectors (i.e. matrix order)
 * @param nev    number of eigenvalues
 */		
int getENCV( int n, int nev )
{
	int ncv = 4 * nev;	
	if (ncv > n)
	{
		ncv = n;
	}
	return ncv;
}

/**
 * Populate EVD arrays from specified eigenvalues.
 *
 * @param nev      number of eigenvalues
 * @param dr       pointer to real eigenvalues.
 * @param reverse  true if the eigenvalues need to be reordered.
 */
void populateEVD( int nev, double *dr, double *evdS, bool reverse )
{
	int i;
	/// POPULATE Eigenvalues
	for (i = 0; i < nev; i++)
	{
		// Do we need to reverse order of eigenvalues?
		// NB: find square root of real eigenvalues
		if( reverse )
		{
			evdS[i] = dr[nev - 1 - i]; 
		}
		else
		{
			evdS[i] = dr[i]; 
		}
	}
}

/**
 * Populate EVD arrays from specified EVD eigenvalues and eigenvectors.
 *
 * @param nev    number of eigenvalues
 * @param dr     pointer to real eigenvalues.
 * @param v      pointer to real eigenvectors.
 */
void populateEVD( Matrix *mm, int nev, double *dr, double *v, double *evdS, DenseMatrix *E, bool reverse )
{
	int n = mm->rows();
	int i, j, eigenvalueIndex;

	// Populate eigenvalues & eigenvectors
	for (i = 0; i < nev; i++)
	{
		// Do we need to reverse order of eigenvalues?
		if( reverse )
		{
			eigenvalueIndex = nev - 1 - i; 
		}
		else
		{
			eigenvalueIndex = i;
		}
		evdS[i] = dr[eigenvalueIndex];
		for (j = 0; j < n; j++)
		{
			// Set E[j][i] = v[i * n + j];
			E->set( j, i, v[eigenvalueIndex * n + j] );
		}
	}

}
