/*
 * Copyright (C) 2006 Derek Greene
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <stdio.h>
#include "Arpack.h"
#include "ArpackSVD.h"
#include "Matrix.h"

// If defined, demand Java to allow access to original arrays not copies
#define USE_CRITICAL 1
#define DEBUG 0

int factorSVD( Matrix *mm, int nev, bool largest, bool computeVectors, double *svdS, double *svdU, double *svdV );
int getNCV( int n, int nev );
void populateSVD( int nev, double *dr, double *svdS, bool reverse );
void populateSVD( Matrix *mm, int nev, double *dr, double *v, double *svdS, DenseMatrix *U, DenseMatrix *V, bool reverse );

// --------------------------------------------------------------------------------------------------------
// JNI METHODS
// --------------------------------------------------------------------------------------------------------

/**
 * Apply ARPACK to a dense matrix.
 */
JNIEXPORT jint JNICALL Java_matrix_decomp_jni_ArpackSVD_factorDense
  (JNIEnv *env, jobject self, 
	jint jK, jint jRows, jint jCols, jdoubleArray jS, jdoubleArray jU, jdoubleArray jV, jdoubleArray jData )
{
	int error;
	double *s, *u, *v, *data;
	jboolean uIsCopy, vIsCopy, sIsCopy, dataIsCopy;
	
	/// Access arrays
#ifdef USE_CRITICAL
	// Handle SVD items
	s = (double *)env->GetPrimitiveArrayCritical( jS, &sIsCopy );
	u = (double *)env->GetPrimitiveArrayCritical( jU, &uIsCopy );
	v = (double *)env->GetPrimitiveArrayCritical( jV, &vIsCopy );
	// Handle matrix specific items
	data = (double *)env->GetPrimitiveArrayCritical( jData, &dataIsCopy );
#else
	// Handle SVD items
	s = env->GetDoubleArrayElements( jS, &sIsCopy );
	u = env->GetDoubleArrayElements( jU, &uIsCopy );
	v = env->GetDoubleArrayElements( jV, &vIsCopy );
	// Handle matrix specific items
	data = env->GetDoubleArrayElements( jData, &dataIsCopy );
#endif

	if( DEBUG ) printf( "Critical=%d sIsCopy=%d uIsCopy=%d vIsCopy=%d dataIsCopy=%d\n", USE_CRITICAL, sIsCopy, uIsCopy, vIsCopy, dataIsCopy );

	/// Create matrix
	Matrix *mm = new DenseMatrix( jRows, jCols, data );

	/// Factor
	error = factorSVD( mm, jK, true, true, s, u, v );

	/// Return the memory to Java
	delete mm;
#ifdef USE_CRITICAL
	env->ReleasePrimitiveArrayCritical( jU, u, JNI_FALSE );
	env->ReleasePrimitiveArrayCritical( jV, v, JNI_FALSE );
	env->ReleasePrimitiveArrayCritical( jS, s, JNI_FALSE );
	env->ReleasePrimitiveArrayCritical( jData, data, JNI_ABORT );
#else
	env->ReleaseDoubleArrayElements( jS, s, JNI_FALSE );
	env->ReleaseDoubleArrayElements( jU, u, JNI_FALSE );
	env->ReleaseDoubleArrayElements( jV, v, JNI_FALSE );
	env->ReleaseDoubleArrayElements( jData, data, JNI_ABORT );
#endif
	return error;
}

/**
 * Apply ARPACK to a sparse compressed-row matrix.
 */
JNIEXPORT jint JNICALL Java_matrix_decomp_jni_ArpackSVD_factorSparseRow
  (JNIEnv *env, jobject self, 
	jint jK, jint jRows, jint jCols, jdoubleArray jS, jdoubleArray jU, jdoubleArray jV, jdoubleArray jData, jintArray jRowPointer, jintArray jUsed, jintArray jColumnIndex)
{
	int error;
	double *s, *u, *v, *data;
	jboolean uIsCopy, vIsCopy, sIsCopy;
	int *rowPointer, *used, *columnIndex;
	jboolean dataIsCopy, rowPointerIsCopy, usedIsCopy, colIndexIsCopy;
	
	/// Access arrays
#ifdef USE_CRITICAL
	// Handle SVD items
	s = (double *)env->GetPrimitiveArrayCritical( jS, &sIsCopy );
	u = (double *)env->GetPrimitiveArrayCritical( jU, &uIsCopy );
	v = (double *)env->GetPrimitiveArrayCritical( jV, &vIsCopy );
	// Handle matrix specific items
	data = (double *)env->GetPrimitiveArrayCritical( jData, &dataIsCopy );
	rowPointer = (int *)env->GetPrimitiveArrayCritical( jRowPointer, &rowPointerIsCopy );
	columnIndex = (int *)env->GetPrimitiveArrayCritical( jColumnIndex, &colIndexIsCopy );
	used = (int *)env->GetPrimitiveArrayCritical( jUsed, &usedIsCopy );
#else
	// Handle SVD items
	s = env->GetDoubleArrayElements( jS, &sIsCopy );
	u = env->GetDoubleArrayElements( jU, &uIsCopy );
	v = env->GetDoubleArrayElements( jV, &vIsCopy );
	// Handle matrix specific items
	data = env->GetDoubleArrayElements( jData, &dataIsCopy );
	rowPointer = env->GetIntArrayElements( jRowPointer, &rowPointerIsCopy );
	used = env->GetIntArrayElements( jUsed, &usedIsCopy );
	columnIndex = env->GetIntArrayElements( jColumnIndex, &colIndexIsCopy );
#endif
	if( DEBUG ) printf( "sIsCopy=%d uIsCopy=%d vIsCopy=%d dataIsCopy=%d ", sIsCopy, uIsCopy, vIsCopy, dataIsCopy );
	if( DEBUG ) printf( "rowPointerIsCopy=%d usedIsCopy=%d colIndexIsCopy=%d\n", rowPointerIsCopy, usedIsCopy, colIndexIsCopy );

	/// Create matrix
	Matrix *mm = new SparseRowMatrix( jRows, jCols, data, rowPointer, used, columnIndex );

	/// Factor
	error = factorSVD( mm, jK, true, true, s, u, v );
	
	/// Return the memory to Java
	delete mm;
#ifdef USE_CRITICAL
	env->ReleasePrimitiveArrayCritical( jU, u, JNI_FALSE );
	env->ReleasePrimitiveArrayCritical( jV, v, JNI_FALSE );
	env->ReleasePrimitiveArrayCritical( jS, s, JNI_FALSE );
	// NB: The JNI_ABORT flag specifies that if you are using a local copy of the array, there is no need to copy back to the real array.
	env->ReleasePrimitiveArrayCritical( jData, data, JNI_ABORT );
	env->ReleasePrimitiveArrayCritical( jRowPointer, rowPointer, JNI_ABORT );
	env->ReleasePrimitiveArrayCritical( jUsed, used, JNI_ABORT );
	env->ReleasePrimitiveArrayCritical( jColumnIndex, columnIndex, JNI_ABORT );
#else
	env->ReleaseDoubleArrayElements( jU, u, JNI_FALSE );
	env->ReleaseDoubleArrayElements( jV, v, JNI_FALSE );
	env->ReleaseDoubleArrayElements( jS, s, JNI_FALSE );
	env->ReleaseDoubleArrayElements( jData, data, JNI_ABORT );
	env->ReleaseIntArrayElements( jRowPointer, rowPointer, JNI_ABORT );
	env->ReleaseIntArrayElements( jUsed, used, JNI_ABORT );
	env->ReleaseIntArrayElements( jColumnIndex, columnIndex, JNI_ABORT );
#endif
	return error;
}

// --------------------------------------------------------------------------------------------------------
// INTERNALS
// --------------------------------------------------------------------------------------------------------


/**
 * This routine uses ARPACK's dsaupd() and dseupd() functions to efficiently 
 * find the SVD of a real rectangular matrix.
 *
 * To use dnaupd(), the only thing that must be supplied in order to use this 
 * routine on your problem is to change the array dimensions 
 * appropriately, to specify WHICH singular values you want to 
 * compute and to supply a the matrix-vector products 
 * 			w <-  Ax 
 * 			y <-  A'w 
 * These are implemented in multRectMatrixVector() and multRectTransposeMatrixVector()
 * respectively.
 *
 * Note that this approach to SVD is just a more efficient way of calculating the
 * eigenvalues of A'*A. The square root of these eigenvalues gives the singular values.
 * The eigenvectors of the matrix give the right singular vectors V. The left singular
 * vectors U are calculated by:
 *
 */
int factorSVD( Matrix *mm, int nev, bool computeLargest, bool computeVectors, double *svdS, double *svdU, double *svdV )
{
	/// Temp variables
	int i, j;

	// Get matrix size
	int rows = mm->rows();
	int cols = mm->columns();
	int n = ( rows > cols ) ? rows : cols;
	
	/// ALLOCATE SPACE
	// Calculate ncv: The largenumber of basis vectors that will be used in the process.
	int ncv = getNCV( n, nev );
	int ldv = n;
	// Space for eigenvectors
	double *v = new double[ldv * ncv];
	// Workspace size
	int lworkl = ncv * (ncv + 8);
	// Space for workspace 
	double *workl = new double[lworkl];
	double *workd = new double[3 * n];
	// Space for eigenvalues (real and imaginary)
	double *d = new double[2 * ncv];
	// Space for residuals
	double *resid = new double[n];
	// NB: Temporary space for transpose multiplication
	double *ax = new double[3 * n];

	// printf( "rows=%d cols=%d n=%d ncv=%d\n", rows, cols, n, ncv );

	/// USER SETTINGS
	// Which eigenvalues to calculate?	
	char* sWhich;
	if( computeLargest )
	{
		// largest algebraic part...
		sWhich = "LA";
	}
	else
	{
		// smallest algebraic part...
		sWhich = "SA";
	}		
	// Specifies whether eigenvectors should be calculated 
	int rvec;		
	if( computeVectors )
	{
		rvec = 1;
	}
	else
	{
		rvec = 0;
	}

	/// INTERNAL SETTINGS
	char bmat[2] = "I";
	// Sets the tolerance
	double tol = 0.0;
	// An array used to pass information to the routines about their functional modes.
	int *iparam = new int[11];
	// Use the exact shift strategy
	iparam[0] = 1;
	// Set  max number of Arnoldi iterations
	iparam[2] = 1000;	
	// Set Mode 1 calculation - standard eigensystem
	iparam[6] = 1; 

	/// STATE
	int ido = 0;
	// Passes convergence information out of the iteration routine.
	int info = 0;
	// Returns info from dneupd
	int ierr;
	// Indicates the locations in the work array workd where the input and output vectors in the
	// callback routine are located
	int *ipntr = new int[14];	
	// TODO: what are these for?
	int *select = new int[ncv];
	double sigma;

	/// PHASE #1: Arnoldi Iteration
	do
	{ 
		// Repeatedly call the routine DNAUPD and take  actions indicated by parameter IDO until
		// either convergence is indicated or maxitr has been exceeded. 
		dsaupd(&ido, bmat, &n, sWhich, &nev, &tol, resid,
			&ncv, v, &ldv, iparam, ipntr, workd, workl,
			&lworkl, &info);
	  
		// Handle state
		if( ( ido == -1 ) || ( ido == 1 ) )
		{
			// find matrix-vector product
			mm->mult( workd + ipntr[0] - 1, ax );
			mm->multTranspose( ax, workd + ipntr[1] - 1 );
		}
	}
	while( ( ido == 1 ) || ( ido == -1 ) );

	// printf( "Info=%d\n", info );
	// Any fatal errors?
	if( info < 0 )
	{
		//// printf( "Error: failure during decomposition, info=%d\n", info );
		return info;
	}
	// Any non-fatal errors
	else if( info == 1 )
	{
		//// printf( "Maximum number of iteration reached\n" );
		return info;
	}
	else if( info > 0 )
	{
		return info;
	}

	/// PHASE #2: Post-Processing

	// Extract the eigenvalues (and eigenvectors if rvec=1)
	dseupd(&rvec, "All", select, d, v, &ldv, &sigma, bmat,
		&n, sWhich, &nev, &tol, resid, &ncv, v, &ldv,
		iparam, ipntr, workd, workl, &lworkl, &ierr);			

	// Any fatal errors?
	if( ierr != 0 )
	{
		// printf( "Error: failure during eigenvalue extraction, info = %d\n", ierr );
		return ( ierr - 1000 );
	}
	
	// Populate SVD arrays
	if( computeVectors )
	{
		DenseMatrix *U = new DenseMatrix( rows, nev, svdU );
		DenseMatrix *V = new DenseMatrix( cols, nev, svdV );
		populateSVD( mm, nev, d, v, svdS, U, V, true );
		delete U, V;
	}
	else
	{
		populateSVD( nev, d, svdS, true );
	}
	
	
	/// TIDY UP
	// deallocate workspace
	delete ax;
	delete workl;
	delete workd;
	delete d;
	delete v;
	delete resid;
	// deallocate settings
	delete iparam;
	// deallocate state
	delete ipntr;
	delete select;

	// Success
	return 0;	
}


/**
 * Calculate the largest number of basis vectors that will be used in the Implicitly Restarted 
 * Arnoldi Process for symmetric and non-symmetric Eigenvalue decomposition.
 *
 * @param n      number of rows for eigenvectors (i.e. matrix order)
 * @param nev    number of eigenvalues
 */		
int getNCV( int n, int nev )
{
	int ncv = 4 * nev;	
	if (ncv > n)
	{
		ncv = n;
	}
	return ncv;
}

/**
 * Populate SVD arrays from specified SVD eigenvalues.
 *
 * @param nev      number of eigenvalues
 * @param dr       pointer to real eigenvalues.
 * @param reverse  true if the eigenvalues need to be reordered.
 */
void populateSVD( int nev, double *dr, double *svdS, bool reverse )
{
	int i;
	/// POPULATE SINGULAR
	for (i = 0; i < nev; i++)
	{
		// Do we need to reverse order of eigenvalues?
		// NB: find square root of real eigenvalues
		if( reverse )
		{
			svdS[i] = sqrt(dr[nev - 1 - i]); 
		}
		else
		{
			svdS[i] = sqrt(dr[i]); 
		}
	}
}

/**
 * Populate SVD arrays from specified SVD eigenvalues and eigenvectors.
 *
 * @param nev    number of eigenvalues
 * @param dr     pointer to real eigenvalues.
 * @param v      pointer to real eigenvectors.
 */
void populateSVD( Matrix *mm, int nev, double *dr, double *v, double *svdS, DenseMatrix *U, DenseMatrix *V, bool reverse )
{
	int rows = mm->rows();
	int cols = mm->columns();
	int i, j, actualEigenIndex;
	// printf( "populateSVD rows=%d cols=%d\n", rows, cols );
	
	/// POPULATE SINGULAR VALUES & RIGHT VECTORS (V) 
	for (i = 0; i < nev; i++)
	{
		// Do we need to reverse order of eigenvalues?
		if( reverse )
		{
			actualEigenIndex = nev - 1 - i; 
		}
		else
		{
			actualEigenIndex = i;
		}
		// NB: find square root of real eigenvalues
		svdS[i] = sqrt(dr[actualEigenIndex]);
		for (j = 0; j < cols; j++)
		{
			V->set( j, i,  v[actualEigenIndex * rows + j] );
		}
	}
	
	/// POPULATE LEFT VECTORS (U)
	double *ax = new double[rows];
	double sv, s, temp;
	// For each eigenvector
	for( i = 0; i < nev; i++ )
	{ 
		// Do we need to reverse order of eigenvalues?
		if( reverse )
		{
			actualEigenIndex = nev - 1 - i; 
		}
		else
		{
			actualEigenIndex = i;
		}
		// zero the vector Ax
		for (j = 0; j < rows; j++)
		{
			ax[j] = 0.0;
		}
		// Multiply matrix by eigenvector
		mm->mult( &v[actualEigenIndex * rows], ax );
		// calculate euclidean norm of Ax
		temp = 0.0;
		for (j = 0; j < rows; j++)
		{
			temp += (ax[j] * ax[j]);
		}
		// avoid NANs
		if( temp != 0.0 )
		{
			temp = 1.0 / sqrt(temp);
			// normalise by euclidean norm
			for (j = 0; j < rows; j++)
			{
				U->set( j, i, (ax[j] * temp) );
			}
		}
	}

	// TIDY UP
	delete [] ax;

}
