/*
 * Copyright (C) 2006 Derek Greene
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifndef ARPACK_MATRIX_H
#define ARPACK_MATRIX_H

#include <stdio.h>

// ---------------------------------------------------------
// Class: Matrix
// ---------------------------------------------------------	

class Matrix
{
	protected:
		int _rows, _cols;
	public:
		Matrix ( int rows, int cols ) : _rows(rows), _cols(cols) {};
		inline int rows() { return _rows; };
		inline int columns() { return _cols; };
		virtual double get( const int row, const int column  ) = 0;
		virtual void set( const int row, const int column, const double value ) = 0;
		virtual void mult( double *in, double *out ) = 0;
		virtual void multTranspose( double *in, double *out ) = 0;

		void print()
		{
			for( int i = 0; i < rows(); i++ )
			{
				for( int j = 0; j < columns(); j++ )
				{
					printf( "%1.3f ", get( i, j  ) );
				}
				printf("\n");
			}	
		};
};

// ---------------------------------------------------------
// Class: DenseMatrix
// ---------------------------------------------------------	

class DenseMatrix : public Matrix
{
	private:
		double *_data;
	public:
		DenseMatrix ( int rows, int cols, double *data ) : Matrix( rows, cols), _data(data) {};

		inline virtual double get( const int row, const int column )
		{
			return _data[row + column * _rows];
		};

		inline virtual void set( const int row, const int column, const double value )
		{
			_data[row + column * _rows] = value;
		};

		inline virtual void mult( double *in, double *out )
		{
			double s;
			int i, j;
			// NB: assume vector size is same as number of cols
			// zero output vector
			for (i = 0; i < _cols; i++)
			{
				out[i] = 0.0;
			}
			// matrix-vector product, as performed in Colt
			for (i = _rows; --i >= 0;)
			{
				s = 0.0;
				for (j = _cols; --j >= 0;)
				{
					s += ( get(i,j) * in[j]);
				}
				out[i] = s;
			}
		};

		inline virtual void multTranspose( double *in, double *out )
		{
			int rowsT, colsT;
			double s;
			int i, j, n;
			// switch size of matrix to transpose
			rowsT = _cols;
			colsT = _rows;
			// NB: assume vector size is same as number of transpose cols
			for (i = 0; i < colsT; i++)
			{
				out[i] = 0.0;
			}
			// matrix-vector product, as performed in Colt
			for (i = rowsT; --i >= 0;)
			{
				s = 0.0;
				for (j = colsT; --j >= 0;)
				{
					s += ( get(j,i) * in[j]);
				}
				out[i] = s;
			}
		};

};

// ---------------------------------------------------------
// Class: SparseRowMatrix
// ---------------------------------------------------------	

/**
 * Compressed Row Storage Matrix wrapper
 */
class SparseRowMatrix : public Matrix
{
	private:
		/** Matrix data */
		double *_data;
		/** Indices to the start of each row */
		int *_rowPointer;
		/** Number of indices in use on each row. */
		int *_used;
		/** Column indices. These are kept sorted within each row. */
		int *_columnIndex;
	
		
	public:
		SparseRowMatrix ( int rows, int cols, double *data, int *rowPointer, int *used, int *columnIndex ) : Matrix( rows, cols), 
			_rowPointer(rowPointer), _used(used), _data(data), _columnIndex(columnIndex) {};

		inline virtual double get( const int row, const int column )
		{
			int index = binarySearch( _columnIndex, column, _rowPointer[row], _rowPointer[row] + _used[row] );
			if (index >= 0) return _data[index];
			return 0.0;
		};

		inline virtual void set( const int row, const int column, const double value )
		{
			// Not supported
		};

		/**
		 * Compressed Sparse Row Matrix - Vector multiple with alpha=1 beta=0.
       * Based on SparseBLAS generator 
		 * http://math.nist.gov/spblas/sourceservice.html
		 */
		inline virtual void mult( double *in, double *out )
		{
			double s;
			const double *ptrData;
			int i,j,jb,je;
			
			// zero output vector
			for (i = 0; i < _cols; i++)
			{
				out[i] = 0.0;
			}
			for (i = 0; i < _rows; ++i)
			{
				s = 0.0;
				for (j = _rowPointer[i]; j < _used[i] + _rowPointer[i]; ++j)
				{
					s += _data[j] * in[_columnIndex[j]];
				}
				out[i] = s;
			}
			
		};

		inline virtual void multTranspose( double *in, double *out )
		{
			int rowsT, colsT;
			double s;
			int i, j, n;
			// switch size of matrix to transpose
			rowsT = _cols;
			colsT = _rows;
			// NB: assume vector size is same as number of transpose cols
			for (i = 0; i < colsT; i++)
			{
				out[i] = 0.0;
			}
			// matrix-vector product, as performed in Colt
			for (i = rowsT; --i >= 0;)
			{
				s = 0.0;
				for (j = colsT; --j >= 0;)
				{
					s += ( get(j,i) * in[j]);
				}
				out[i] = s;
			}
		};

		inline virtual int binarySearch( int *index, int key, int begin, int end )
		{
			end--;
			while (begin <= end)
			{
				int mid = (end + begin) >> 1;
				if (index[mid] < key)
				{
					begin = mid + 1;
				}
				else if (index[mid] > key)
				{
					end = mid - 1;
				}
				else
				{
					return mid;
				}
			}
			return -1;
		};


};

#endif // ARPACK_MATRIX_H

